"""Source code analyzer for chalice app.

The main point of this module is to analyze your source code
and track which AWS API calls you make.

We can then use this information to create IAM policies
automatically for you.

How it Works
============

This is basically a simplified abstract interpreter.
The type inference is greatly simplified because
we're only interested in boto3 client types.
In a nutshell:

* Create an AST and symbol table from the source code.
* Interpret the AST and track boto3 types.  This is governed
  by a few simple rules.
* Propagate inferred boto3 types as much as possible.  Most of
  the basic stuff is handled, for example:

      * ``x = y`` if y is a boto3 type, so is x.
      * ``a :: (x -> y), where y is a boto3 type, then given ``b = a()``,
        b is of type y.
      * Map inferred types across function params and return types.

At the end of the analysis, a final walk is performed to collect any
node of type ``Boto3ClientMethodCallType``.  This represents an
API call being made.  This also lets you be selective about which
API calls you care about.  For example, if you want only want to see
which API calls happen in a particular function, only walk that
particular ``FunctionDef`` node.

"""
import ast
import symtable

from typing import Dict, Set, Any, Optional, List, Union, cast  # noqa


APICallT = Dict[str, Set[str]]
OptASTSet = Optional[Set[ast.AST]]
ComprehensionNode = Union[ast.DictComp, ast.GeneratorExp, ast.ListComp]


def get_client_calls(source_code):
    # type: (str) -> APICallT
    """Return all clients calls made in provided source code.

    :returns: A dict of service_name -> set([client calls]).
        Example: {"s3": set(["list_objects", "create_bucket"]),
                  "dynamodb": set(["describe_table"])}
    """
    parsed = parse_code(source_code)
    t = SymbolTableTypeInfer(parsed)
    binder = t.bind_types()
    collector = APICallCollector(binder)
    api_calls = collector.collect_api_calls(parsed.parsed_ast)
    return api_calls


def get_client_calls_for_app(source_code):
    # type: (str) -> APICallT
    """Return client calls for a chalice app.

    This is similar to ``get_client_calls`` except it will
    automatically traverse into chalice views with the assumption
    that they will be called.

    """
    parsed = parse_code(source_code)
    t = AppViewTypeInfer(parsed)
    binder = t.bind_types()
    collector = APICallCollector(binder)
    api_calls = collector.collect_api_calls(parsed.parsed_ast)
    return api_calls


def parse_code(source_code, filename='app.py'):
    # type: (str, str) -> ParsedCode
    parsed = ast.parse(source_code, filename)
    table = symtable.symtable(source_code, filename, 'exec')
    return ParsedCode(parsed, ChainedSymbolTable(table, table))


class BaseType(object):
    def __repr__(self):
        # type: () -> str
        return "%s()" % self.__class__.__name__

    def __eq__(self, other):
        # type: (Any) -> bool
        return isinstance(other, self.__class__)


# The next 5 classes are used to track the
# components needed to create a boto3 client.
# While we really only care about boto3 clients we need
# to track all the types it takes to get there:
#
# import boto3          <--- bind "boto3" as the boto3 module type
# c = boto.client       <--- bind "c" as the boto3 create client type
# s3 = c('s3')          <--- bind 's3' as the boto3 client type, subtype 's3'.
# m = s3.list_objects   <--- bind as API call 's3', 'list_objects'
# r = m()               <--- bind as API call invoked (what we care about).
#
# That way we can handle (in addition to the case above) things like:
# import boto3; boto3.client('s3').list_objects()
# import boto3; s3 = boto3.client('s3'); s3.list_objects()
class Boto3ModuleType(BaseType):
    pass


class Boto3CreateClientType(BaseType):
    pass


class Boto3ClientType(BaseType):
    def __init__(self, service_name):
        # type: (str) -> None
        #: The name of the AWS service, e.g. 's3'.
        self.service_name = service_name

    def __eq__(self, other):
        # type: (Any) -> bool
        # NOTE: We can't use self.__class__ because of a mypy bug:
        # https://github.com/python/mypy/issues/3061
        # We can change this back once that bug is fixed.
        if not isinstance(other, Boto3ClientType):
            return False
        return self.service_name == other.service_name

    def __repr__(self):
        # type: () -> str
        return "%s(%s)" % (self.__class__.__name__, self.service_name)


class Boto3ClientMethodType(BaseType):
    def __init__(self, service_name, method_name):
        # type: (str, str) -> None
        self.service_name = service_name
        self.method_name = method_name

    def __eq__(self, other):
        # type: (Any) -> bool
        if self.__class__ != other.__class__:
            return False
        return (
            self.service_name == other.service_name and
            self.method_name == other.method_name)

    def __repr__(self):
        # type: () -> str
        return "%s(%s, %s)" % (
            self.__class__.__name__,
            self.service_name,
            self.method_name
        )


class Boto3ClientMethodCallType(Boto3ClientMethodType):
    pass


class TypedSymbol(symtable.Symbol):
    inferred_type = None  # type: Any
    ast_node = None  # type: ast.AST


class FunctionType(BaseType):
    def __init__(self, return_type):
        # type: (Any) -> None
        self.return_type = return_type

    def __eq__(self, other):
        # type: (Any) -> bool
        if self.__class__ != other.__class__:
            return False
        return self.return_type == other.return_type

    def __repr__(self):
        # type: () -> str
        return "%s(%s)" % (
            self.__class__.__name__,
            self.return_type,
        )


class StringLiteral(object):
    def __init__(self, value):
        # type: (str) -> None
        self.value = value


class ParsedCode(object):
    def __init__(self, parsed_ast, symbol_table):
        # type: (ast.AST, ChainedSymbolTable) -> None
        self.parsed_ast = parsed_ast
        self.symbol_table = symbol_table


class APICallCollector(ast.NodeVisitor):
    """Traverse a given AST and look for any inferred API call types.

    This visitor assumes you've ran type inference on the AST.
    It will search through the AST and collect any API calls.
    """
    def __init__(self, binder):
        # type: (TypeBinder) -> None
        self.api_calls = {}  # type: APICallT
        self._binder = binder

    def collect_api_calls(self, node):
        # type: (ast.AST) -> APICallT
        self.visit(node)
        return self.api_calls

    def visit(self, node):
        # type: (ast.AST) -> None
        inferred_type = self._binder.get_type_for_node(node)
        if isinstance(inferred_type, Boto3ClientMethodCallType):
            self.api_calls.setdefault(inferred_type.service_name, set()).add(
                inferred_type.method_name)
        ast.NodeVisitor.visit(self, node)


class ChainedSymbolTable(object):
    def __init__(self, local_table, global_table):
        # type: (symtable.SymbolTable, symtable.SymbolTable) -> None
        # If you're in the module scope, then pass in
        # the same symbol table for local and global.
        self._local_table = local_table
        self._global_table = global_table

    def new_sub_table(self, local_table):
        # type: (symtable.SymbolTable) -> ChainedSymbolTable
        # Create a new symbol table using this instances
        # local table as the new global table and the passed
        # in local table as the new local table.
        return self.__class__(local_table, self._local_table)

    def get_inferred_type(self, name):
        # type: (str) -> Any
        # Given a symbol name, check whether a type
        # has been inferred.
        # The stdlib symtable will already fall back to
        # global scope if necessary.
        symbol = self._local_table.lookup(name)
        if symbol.is_global():
            try:
                global_symbol = self._global_table.lookup(name)
            except KeyError:
                # It's not an error if a symbol.is_global()
                # but is not in our "_global_table", because
                # we're not considering the builtin scope.
                # In this case we just say that there is no
                # type we've inferred.
                return None
            return getattr(global_symbol, 'inferred_type', None)
        return getattr(symbol, 'inferred_type', None)

    def set_inferred_type(self, name, inferred_type):
        # type: (str, Any) -> None
        symbol = cast(TypedSymbol, self._local_table.lookup(name))
        symbol.inferred_type = inferred_type

    def lookup_sub_namespace(self, name, lineno=None):
        # type: (str, Optional[int]) -> ChainedSymbolTable
        for child in self._local_table.get_children():
            if child.get_name() == name:
                if lineno is not None:
                    if child.get_lineno() == lineno:
                        return self.__class__(child, self._local_table)
                else:
                    return self.__class__(child, self._local_table)
        for child in self._global_table.get_children():
            if child.get_name() == name:
                return self.__class__(child, self._global_table)
        raise ValueError("Unknown symbol name: %s" % name)

    def get_sub_namespaces(self):
        # type: () -> List[symtable.SymbolTable]
        return self._local_table.get_children()

    def get_name(self):
        # type: () -> str
        return self._local_table.get_name()

    def get_symbols(self):
        # type: () -> List[symtable.Symbol]
        return self._local_table.get_symbols()

    def register_ast_node_for_symbol(self, name, node):
        # type: (str, ast.AST) -> None
        symbol = cast(TypedSymbol, self._local_table.lookup(name))
        symbol.ast_node = node

    def lookup_ast_node_for_symbol(self, name):
        # type: (str) -> ast.AST
        symbol = self._local_table.lookup(name)
        if symbol.is_global():
            symbol = self._global_table.lookup(name)
        try:
            return cast(TypedSymbol, symbol).ast_node
        except AttributeError:
            raise ValueError(
                "No AST node registered for symbol: %s" % name)

    def has_ast_node_for_symbol(self, name):
        # type: (str) -> bool
        try:
            self.lookup_ast_node_for_symbol(name)
            return True
        except (ValueError, KeyError):
            return False


class TypeBinder(object):

    def __init__(self):
        # type: () -> None
        self._node_to_type = {}  # type: Dict[ast.AST, Any]

    def get_type_for_node(self, node):
        # type: (Any) -> Any
        return self._node_to_type.get(node)

    def set_type_for_node(self, node, inferred_type):
        # type: (Any, Any) -> None
        self._node_to_type[node] = inferred_type


class SymbolTableTypeInfer(ast.NodeVisitor):
    _SDK_PACKAGE = 'boto3'
    _CREATE_CLIENT = 'client'

    def __init__(self, parsed_code, binder=None, visited=None):
        # type: (ParsedCode, Optional[TypeBinder], OptASTSet) -> None
        self._symbol_table = parsed_code.symbol_table
        self._current_ast_namespace = parsed_code.parsed_ast
        self._node_inference = {}  # type: Dict[ast.AST, Any]
        if binder is None:
            binder = TypeBinder()
        if visited is None:
            visited = set()
        self._binder = binder
        self._visited = visited

    def bind_types(self):
        # type: () -> TypeBinder
        self.visit(self._current_ast_namespace)
        return self._binder

    def known_types(self, scope_name=None):
        # type: (Optional[str]) -> Dict[str, Any]
        table = None
        if scope_name is None:
            table = self._symbol_table
        else:
            table = self._symbol_table.lookup_sub_namespace(scope_name)
        return {
            s.get_name(): cast(TypedSymbol, s).inferred_type
            for s in table.get_symbols()
            if hasattr(s, 'inferred_type') and
            cast(TypedSymbol, s).inferred_type is not None and
            s.is_local()
        }

    def _set_inferred_type_for_name(self, name, inferred_type):
        # type: (str, Any) -> None
        self._symbol_table.set_inferred_type(name, inferred_type)

    def _set_inferred_type_for_node(self, node, inferred_type):
        # type: (Any, Any) -> None
        self._binder.set_type_for_node(node, inferred_type)

    def _get_inferred_type_for_node(self, node):
        # type: (Any) -> Any
        return self._binder.get_type_for_node(node)

    def _new_inference_scope(self, parsed_code, binder, visited):
        # type: (ParsedCode, TypeBinder, Set[ast.AST]) -> SymbolTableTypeInfer
        instance = self.__class__(parsed_code, binder, visited)
        return instance

    def visit_Import(self, node):
        # type: (ast.Import) -> None
        for child in node.names:
            if isinstance(child, ast.alias):
                import_name = child.name
                if import_name == self._SDK_PACKAGE:
                    self._set_inferred_type_for_name(
                        import_name, Boto3ModuleType())
        self.generic_visit(node)

    def visit_Name(self, node):
        # type: (ast.Name) -> None
        self._set_inferred_type_for_node(
            node,
            self._symbol_table.get_inferred_type(node.id)
        )
        self.generic_visit(node)

    def visit_Assign(self, node):
        # type: (ast.Assign) -> None
        # The LHS gets the inferred type of the RHS.
        # We do this post-traversal to let the type inference
        # run on the children first.
        self.generic_visit(node)
        rhs_inferred_type = self._get_inferred_type_for_node(node.value)
        if rhs_inferred_type is None:
            # Special casing assignment to a string literal.
            if isinstance(node.value, ast.Str):
                rhs_inferred_type = StringLiteral(node.value.s)
                self._set_inferred_type_for_node(node.value, rhs_inferred_type)
        for t in node.targets:
            if isinstance(t, ast.Name):
                self._symbol_table.set_inferred_type(t.id, rhs_inferred_type)
                self._set_inferred_type_for_node(node, rhs_inferred_type)

    def visit_Attribute(self, node):
        # type: (ast.Attribute) -> None
        self.generic_visit(node)
        lhs_inferred_type = self._get_inferred_type_for_node(node.value)
        if lhs_inferred_type is None:
            return
        elif lhs_inferred_type == Boto3ModuleType():
            # Check for attributes such as boto3.client.
            if node.attr == self._CREATE_CLIENT:
                # This is a "boto3.client" attribute.
                self._set_inferred_type_for_node(node, Boto3CreateClientType())
        elif isinstance(lhs_inferred_type, Boto3ClientType):
            self._set_inferred_type_for_node(
                node,
                Boto3ClientMethodType(
                    lhs_inferred_type.service_name,
                    node.attr
                )
            )

    def visit_Call(self, node):
        # type: (ast.Call) -> None
        self.generic_visit(node)
        # func -> Node that's being called
        # args -> Arguments being passed.
        inferred_func_type = self._get_inferred_type_for_node(node.func)
        if inferred_func_type == Boto3CreateClientType():
            # e_0 : B3CCT -> B3CT[S]
            # e_1 : S str which is a service name
            # e_0(e_1) : B3CT[e_1]
            if len(node.args) >= 1:
                service_arg = node.args[0]
                if isinstance(service_arg, ast.Str):
                    self._set_inferred_type_for_node(
                        node, Boto3ClientType(service_arg.s))
                elif isinstance(self._get_inferred_type_for_node(service_arg),
                                StringLiteral):
                    sub_type = self._get_inferred_type_for_node(service_arg)
                    inferred_type = Boto3ClientType(sub_type.value)
                    self._set_inferred_type_for_node(node, inferred_type)
        elif isinstance(inferred_func_type, Boto3ClientMethodType):
            self._set_inferred_type_for_node(
                node,
                Boto3ClientMethodCallType(
                    inferred_func_type.service_name,
                    inferred_func_type.method_name
                )
            )
        elif isinstance(inferred_func_type, FunctionType):
            self._set_inferred_type_for_node(
                node, inferred_func_type.return_type)
        elif isinstance(node.func, ast.Name) and \
                self._symbol_table.has_ast_node_for_symbol(node.func.id):
            if node not in self._visited:
                self._visited.add(node)
                self._infer_function_call(node)

    def visit_Lambda(self, node):
        # type: (ast.Lambda) -> None
        # Lambda is going to be a bit tricky because
        # there's a new child namespace (via .get_children()),
        # but it's not something that will show up in the
        # current symbol table via .lookup().
        # For now, we're going to ignore lambda expressions.
        pass

    def _infer_function_call(self, node):
        # type: (Any) -> None
        # Here we're calling a function we haven't analyzed
        # yet.  We're first going to analyze the function.
        # This will set the inferred_type on the FunctionDef
        # node.
        # If we get a FunctionType as the inferred type of the
        # function, then we know that the inferred type for
        # calling the function is the .return_type type.
        function_name = node.func.id
        sub_table = self._symbol_table.lookup_sub_namespace(
            function_name, node.lineno)
        ast_node = self._symbol_table.lookup_ast_node_for_symbol(
            function_name)

        self._map_function_params(sub_table, node, ast_node)

        child_infer = self._new_inference_scope(
            ParsedCode(ast_node, sub_table), self._binder, self._visited)
        child_infer.bind_types()
        inferred_func_type = self._get_inferred_type_for_node(ast_node)
        self._symbol_table.set_inferred_type(function_name, inferred_func_type)
        # And finally the result of this Call() node will be
        # the return type from the function we just analyzed.
        if isinstance(inferred_func_type, FunctionType):
            self._set_inferred_type_for_node(
                node, inferred_func_type.return_type)

    def _map_function_params(self, sub_table, node, def_node):
        # type: (ChainedSymbolTable, Any, Any) -> None
        # TODO: Handle the full calling syntax, kwargs, stargs, etc.
        #       Right now we just handle positional args.
        defined_args = def_node.args
        for arg, defined in zip(node.args, defined_args.args):
            inferred_type = self._get_inferred_type_for_node(arg)
            if inferred_type is not None:
                name = self._get_name(defined)
                sub_table.set_inferred_type(name, inferred_type)

    def _get_name(self, node):
        # type: (Any) -> str
        try:
            return getattr(node, 'id')
        except AttributeError:
            return getattr(node, 'arg')

    def visit_FunctionDef(self, node):
        # type: (ast.FunctionDef) -> None
        if node.name == self._symbol_table.get_name():
            # Not using generic_visit() because we don't want to
            # visit the decorator_list attr.
            for child in node.body:
                self.visit(child)
        else:
            self._symbol_table.register_ast_node_for_symbol(node.name, node)

    def visit_AsyncFunctionDef(self, node):
        # type: (ast.AsyncFunctionDef) -> None
        # this type is actually wrong but we can't use the actual type as it's
        # not available in python 2
        converted = cast(ast.FunctionDef, node)
        self.visit_FunctionDef(converted)

    def visit_ClassDef(self, node):
        # type: (ast.ClassDef) -> None
        # Not implemented yet.  We want to ensure we don't
        # traverse into the class body for now.
        return

    def visit_DictComp(self, node):
        # type: (ast.DictComp) -> None
        self._handle_comprehension(node, 'dictcomp')

    def visit_Return(self, node):
        # type: (Any) -> None
        self.generic_visit(node)
        inferred_type = self._get_inferred_type_for_node(node.value)
        if inferred_type is not None:
            self._set_inferred_type_for_node(node, inferred_type)
            # We're making a pretty big assumption there's one return
            # type per function.  Will likely need to come back to this.
            inferred_func_type = FunctionType(inferred_type)
            self._set_inferred_type_for_node(self._current_ast_namespace,
                                             inferred_func_type)

    def visit_ListComp(self, node):
        # type: (ast.ListComp) -> None
        # 'listcomp' is the string literal used by python
        # to creating the SymbolTable for the corresponding
        # list comp function.
        self._handle_comprehension(node, 'listcomp')

    def visit_GeneratorExp(self, node):
        # type: (ast.GeneratorExp) -> None
        # Generator expressions are an interesting case.
        # They create a new sub scope, but they're not
        # explicitly named.  Python just creates a table
        # with the name "genexpr".
        self._handle_comprehension(node, 'genexpr')

    def _visit_first_comprehension_generator(self, node):
        # type: (ComprehensionNode) -> None
        if node.generators:
            # first generator's iterator is visited in the current scope
            first_generator = node.generators[0]
            self.visit(first_generator.iter)

    def _collect_comprehension_children(self, node):
        # type: (ComprehensionNode) -> List[ast.expr]
        if isinstance(node, ast.DictComp):
            # dict comprehensions have two values to be checked
            child_nodes = [node.key, node.value]
        else:
            child_nodes = [node.elt]

        if node.generators:
            first_generator = node.generators[0]
            child_nodes.append(first_generator.target)
            for if_expr in first_generator.ifs:
                child_nodes.append(if_expr)

        for generator in node.generators[1:]:
            # rest need to be visited in the child scope
            child_nodes.append(generator.iter)
            child_nodes.append(generator.target)
            for if_expr in generator.ifs:
                child_nodes.append(if_expr)
        return child_nodes

    def _visit_comprehension_children(self, node, comprehension_type):
        # type: (ComprehensionNode, str) -> None
        child_nodes = self._collect_comprehension_children(node)
        child_scope = self._get_matching_sub_namespace(comprehension_type,
                                                       node.lineno)
        if child_scope is None:
            # In Python 2 there's no child scope for list comp
            # Or we failed to locate the child scope, this happens in Python 2
            # when there are multiple comprehensions of the same type in the
            # same scope. The line number trick doesn't work as Python 2 always
            # passes line number 0, make a best effort
            for child_node in child_nodes:
                try:
                    self.visit(child_node)
                except KeyError:
                    pass
            return
        for child_node in child_nodes:
            # visit sub expressions in the child scope
            child_table = self._symbol_table.new_sub_table(child_scope)
            child_infer = self._new_inference_scope(
                ParsedCode(child_node, child_table),
                self._binder, self._visited)
            child_infer.bind_types()

    def _handle_comprehension(self, node, comprehension_type):
        # type: (ComprehensionNode, str) -> None
        self._visit_first_comprehension_generator(node)
        self._visit_comprehension_children(node, comprehension_type)

    def _get_matching_sub_namespace(self, name, lineno):
        # type: (str, int) -> Optional[symtable.SymbolTable]
        namespaces = [t for t in self._symbol_table.get_sub_namespaces()
                      if t.get_name() == name]
        if len(namespaces) == 1:
            # if there's only one match for the name, return it
            return namespaces[0]
        for namespace in namespaces:
            # otherwise disambiguate by using the line number
            if namespace.get_lineno() == lineno:
                return namespace
        return None

    def visit(self, node):
        # type: (Any) -> None
        return ast.NodeVisitor.visit(self, node)


class AppViewTypeInfer(ast.NodeVisitor):
    _CHALICE_DECORATORS = [
        'route', 'authorizer', 'lambda_function',
        'schedule', 'on_s3_event', 'on_sns_message',
        'on_sqs_message', 'on_ws_connect', 'on_ws_message',
        'on_ws_disconnect',
    ]

    def __init__(self, parsed_code):
        # type: (ParsedCode) -> None
        self._binder = TypeBinder()
        self._visited = set()  # type: Set[ast.AST]
        self._parsed_code = parsed_code
        self._type_infer = SymbolTableTypeInfer(
            self._parsed_code, self._binder, self._visited)

    def bind_types(self):
        # type: () -> TypeBinder
        self._type_infer.bind_types()
        self.visit(self._parsed_code.parsed_ast)
        return self._binder

    def visit_FunctionDef(self, node):
        # type: (ast.FunctionDef) -> None
        if self._is_chalice_view(node):
            sub_table = self._parsed_code.symbol_table.lookup_sub_namespace(
                node.name, node.lineno)
            child_infer = SymbolTableTypeInfer(
                ParsedCode(node, sub_table), self._binder, self._visited)
            child_infer.bind_types()

    def _is_chalice_view(self, node):
        # type: (ast.FunctionDef) -> bool
        # We can certainly improve on this, but this check is more
        # of a heuristic for the time being.  The ideal way to do this
        # is to infer the Chalice type and ensure the function is
        # decorated with the Chalice type's route() method.
        decorator_list = node.decorator_list
        if not decorator_list:
            return False
        for decorator in decorator_list:
            if isinstance(decorator, ast.Call) and \
                    isinstance(decorator.func, ast.Attribute):
                if decorator.func.attr in self._CHALICE_DECORATORS:
                    return True
        return False



"""New project generation.

How it Works
============

Project template are placed with the ./chalice/templates directory.  Each
directory corresponds to a single template.  The name of the directory has
the structure ``0123-template-name``, where the first part consists of a four
digit number followed by a ``-``, then the name of the template.  The leading
number is not exposed externally to the user and is used solely for sorting
purposes so we can display the project templates in the order we prefer.
The template name is the name that's used in the ``--project-type`` value
for the ``new-project`` command.

Each template can have a ``DESCRIPTION`` file that contains a short
description of the project template.  This will be used as the display
value instead of the project type key if this file is available.  The
``DESCRIPTION`` file is not copied over when generating the new project
files.

There's basic support for templating values.  This allows you to write
templates with placeholder values that are filled in during project
generation time.  These values are denoted via ``{{template_var}}``.
The following keys are supported:

* ``app_name`` - The name of the project.
* ``chalice_version`` - The current version of chalice generating the project.

"""
from __future__ import print_function
import os
import re
import json
import fnmatch
from typing import Optional, Dict, Any, Iterator, Tuple, Match, List  # noqa

import inquirer
from attr import attrs, attrib

from chalice.constants import WELCOME_PROMPT
from chalice.utils import OSUtils
from chalice.app import __version__ as chalice_version


TEMPLATES_DIR = os.path.join(
    os.path.dirname(
        os.path.dirname(os.path.abspath(__file__))),
    'templates'
)
VAR_REF_REGEX = r'{{(.*?)}}'
IGNORE_FILES = ['metadata.json', '*.pyc']


class BadTemplateError(Exception):
    pass


def create_new_project_skeleton(project_name, project_type='legacy'):
    # type: (str, Optional[str]) -> None
    osutils = OSUtils()
    all_projects = list_available_projects(TEMPLATES_DIR, osutils)
    project = [p for p in all_projects if p.key == project_type][0]
    template_kwargs = {
        'app_name': project_name,
        'chalice_version': chalice_version,
    }
    project_creator = ProjectCreator(osutils)
    project_creator.create_new_project(
        os.path.join(TEMPLATES_DIR, project.dirname),
        project_name,
        template_kwargs=template_kwargs,
    )


@attrs
class ProjectTemplate(object):
    dirname = attrib()    # type: str
    metadata = attrib()   # type: Dict[str, Any]
    key = attrib()        # type: str

    @property
    def description(self):
        # type: () -> str
        # Pylint doesn't understand the attrs types.
        # pylint: disable=no-member
        return self.metadata.get('description', self.key)


class ProjectCreator(object):
    def __init__(self, osutils=None):
        # type: (Optional[OSUtils]) -> None
        if osutils is None:
            osutils = OSUtils()
        self._osutils = osutils

    def create_new_project(self, source_dir, destination_dir, template_kwargs):
        # type: (str, str, Dict[str, Any]) -> None
        for full_src_path, full_dst_path in self._iter_files(source_dir,
                                                             destination_dir):
            dest_dir = self._osutils.dirname(full_dst_path)
            if not self._osutils.directory_exists(dest_dir):
                self._osutils.makedirs(dest_dir)
            contents = self._osutils.get_file_contents(full_src_path,
                                                       binary=False)
            templated_contents = get_templated_content(
                contents, template_kwargs)
            self._osutils.set_file_contents(
                full_dst_path, templated_contents, binary=False)

    def _iter_files(self, source_dir, destination_dir):
        # type: (str, str) -> Iterator[Tuple[str, str]]
        for rootdir, _, filenames in self._osutils.walk(source_dir):
            for filename in filenames:
                if self._should_ignore(filename):
                    continue
                full_src_path = os.path.join(rootdir, filename)
                # The starting index needs `+ 1` to account for the
                # trailing `/` char (e.g. foo/bar -> foo/bar/).
                full_dst_path = os.path.join(
                    destination_dir, full_src_path[len(source_dir) + 1:])
                yield full_src_path, full_dst_path

    def _should_ignore(self, filename):
        # type: (str) -> bool
        for ignore in IGNORE_FILES:
            if fnmatch.fnmatch(filename, ignore):
                return True
        return False


def get_templated_content(contents, template_kwargs):
    # type: (str, Dict[str, Any]) -> str
    def lookup_var(match):
        # type: (Match) -> str
        var_name = match.group(1)
        try:
            return template_kwargs[var_name]
        except KeyError:
            raise BadTemplateError(
                "Bad template, referenced template var that does not "
                "exist: '%s', for template contents:\n%s" % (var_name,
                                                             contents)
            )

    new_contents = re.sub(VAR_REF_REGEX, lookup_var, contents)
    return new_contents


def list_available_projects(templates_dir, osutils):
    # type: (str, OSUtils) -> List[ProjectTemplate]
    projects = []
    for dirname in sorted(osutils.get_directory_contents(templates_dir)):
        filename = osutils.joinpath(templates_dir, dirname, 'metadata.json')
        metadata = json.loads(osutils.get_file_contents(filename, False))
        key = dirname.split('-', 1)[1]
        projects.append(ProjectTemplate(dirname, metadata, key=key))
    return projects


def getting_started_prompt():
    # type: () -> Dict[str, Any]
    print(WELCOME_PROMPT)
    projects = list_available_projects(TEMPLATES_DIR, OSUtils())
    questions = [
        inquirer.Text('project_name', message='Enter the project name'),
        inquirer.List('project_type', message='Select your project type',
                      choices=[(p.description, p.key) for p in projects])
    ]
    answers = inquirer.prompt(questions)
    return answers



from chalice.test import Client
from app import app


def test_s3_handler():
    with Client(app) as client:
        event = client.events.generate_s3_event(
            bucket='mybucket', key='mykey')
        client.lambda_.invoke('s3_handler', event)



# pylint: disable=too-many-lines

import copy
import json
import os
import re

import six
from typing import Any, Optional, Dict, List, Set, Union  # noqa
from typing import cast

import yaml
from yaml.scanner import ScannerError
from yaml.nodes import Node, ScalarNode, SequenceNode

from chalice.deploy.swagger import (
    CFNSwaggerGenerator, TerraformSwaggerGenerator)
from chalice.utils import (
    OSUtils, UI, serialize_to_json, to_cfn_resource_name
)
from chalice.awsclient import TypedAWSClient # noqa
from chalice.config import Config  # noqa
from chalice.deploy import models
from chalice.deploy.appgraph import ApplicationGraphBuilder, DependencyBuilder
from chalice.deploy.deployer import BuildStage  # noqa
from chalice.deploy.deployer import create_build_stage


def create_app_packager(
        config, options, package_format='cloudformation',
        template_format='json', merge_template=None):
    # type: (Config, PackageOptions, str, str, Optional[str]) -> AppPackager
    osutils = OSUtils()
    ui = UI()
    application_builder = ApplicationGraphBuilder()
    deps_builder = DependencyBuilder()
    post_processors = []  # type: List[TemplatePostProcessor]
    generator = None  # type: Union[None, TemplateGenerator]

    template_serializer = cast(TemplateSerializer, JSONTemplateSerializer())
    if package_format == 'cloudformation':
        build_stage = create_build_stage(
            osutils, ui, CFNSwaggerGenerator(), config)
        use_yaml_serializer = template_format == 'yaml'
        if merge_template is not None and \
                YAMLTemplateSerializer.is_yaml_template(merge_template):
            # Automatically switch the serializer to yaml if they specify
            # a yaml template to merge, regardless of what template format
            # they specify.
            use_yaml_serializer = True
        if use_yaml_serializer:
            template_serializer = YAMLTemplateSerializer()
        post_processors.extend([
            SAMCodeLocationPostProcessor(osutils=osutils),
            TemplateMergePostProcessor(
                osutils=osutils,
                merger=TemplateDeepMerger(),
                template_serializer=template_serializer,
                merge_template=merge_template)])
        generator = SAMTemplateGenerator(config, options)
    else:
        build_stage = create_build_stage(
            osutils, ui, TerraformSwaggerGenerator(), config)
        generator = TerraformGenerator(config, options)
        post_processors.append(
            TerraformCodeLocationPostProcessor(osutils=osutils))

    resource_builder = ResourceBuilder(
        application_builder, deps_builder, build_stage)

    return AppPackager(
        generator,
        resource_builder,
        CompositePostProcessor(post_processors),
        template_serializer,
        osutils)


class UnsupportedFeatureError(Exception):
    pass


class DuplicateResourceNameError(Exception):
    pass


class PackageOptions(object):
    def __init__(self, client):
        # type: (TypedAWSClient) -> None
        self._client = client  # type: TypedAWSClient

    def service_principal(self, service):
        # type: (str) -> str
        dns_suffix = self._client.endpoint_dns_suffix(service,
                                                      self._client.region_name)
        return self._client.service_principal(service,
                                              self._client.region_name,
                                              dns_suffix)


class ResourceBuilder(object):
    def __init__(self,
                 application_builder,  # type: ApplicationGraphBuilder
                 deps_builder,         # type: DependencyBuilder
                 build_stage,          # type: BuildStage
                 ):
        # type: (...) -> None
        self._application_builder = application_builder
        self._deps_builder = deps_builder
        self._build_stage = build_stage

    def construct_resources(self, config, chalice_stage_name):
        # type: (Config, str) -> List[models.Model]
        application = self._application_builder.build(
            config, chalice_stage_name)
        resources = self._deps_builder.build_dependencies(application)
        self._build_stage.execute(config, resources)
        # Rebuild dependencies in case the build stage modified
        # the application graph.
        resources = self._deps_builder.build_dependencies(application)
        return resources


class TemplateGenerator(object):

    template_file = None  # type: str

    def __init__(self, config, options):
        # type: (Config, PackageOptions) -> None
        self._config = config
        self._options = options

    def dispatch(self, resource, template):
        # type: (models.Model, Dict[str, Any]) -> None
        name = '_generate_%s' % resource.__class__.__name__.lower()
        handler = getattr(self, name, self._default)
        handler(resource, template)

    def generate(self, resources):
        # type: (List[models.Model]) -> Dict[str, Any]
        raise NotImplementedError()

    def _generate_filebasediampolicy(self, resource, template):
        # type: (models.FileBasedIAMPolicy, Dict[str, Any]) -> None
        pass

    def _generate_autogeniampolicy(self, resource, template):
        # type: (models.AutoGenIAMPolicy, Dict[str, Any]) -> None
        pass

    def _generate_deploymentpackage(self, resource, template):
        # type: (models.DeploymentPackage, Dict[str, Any]) -> None
        pass

    def _generate_precreatediamrole(self, resource, template):
        # type: (models.PreCreatedIAMRole, Dict[str, Any]) -> None
        pass

    def _default(self, resource, template):
        # type: (models.Model, Dict[str, Any]) -> None
        raise UnsupportedFeatureError(resource)


class SAMTemplateGenerator(TemplateGenerator):
    _BASE_TEMPLATE = {
        'AWSTemplateFormatVersion': '2010-09-09',
        'Transform': 'AWS::Serverless-2016-10-31',
        'Outputs': {},
        'Resources': {},
    }

    template_file = "sam"

    def __init__(self, config, options):
        # type: (Config, PackageOptions) -> None
        super(SAMTemplateGenerator, self).__init__(config, options)
        self._seen_names = set([])  # type: Set[str]
        self._chalice_layer = ""

    def generate(self, resources):
        # type: (List[models.Model]) -> Dict[str, Any]
        template = copy.deepcopy(self._BASE_TEMPLATE)
        self._seen_names.clear()
        for resource in resources:
            self.dispatch(resource, template)
        return template

    def _generate_lambdalayer(self, resource, template):
        # type: (models.LambdaLayer, Dict[str, Any]) -> None
        layer = to_cfn_resource_name(
            resource.resource_name)
        template['Resources'][layer] = {
            "Type": "AWS::Serverless::LayerVersion",
            "Properties": {
                "CompatibleRuntimes": [resource.runtime],
                "ContentUri": resource.deployment_package.filename,
                "LayerName": resource.layer_name
            }
        }
        self._chalice_layer = layer

    def _generate_scheduledevent(self, resource, template):
        # type: (models.ScheduledEvent, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        event_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        function_cfn['Properties']['Events'] = {
            event_cfn_name: {
                'Type': 'Schedule',
                'Properties': {
                    'Schedule': resource.schedule_expression,
                }
            }
        }

    def _generate_cloudwatchevent(self, resource, template):
        # type: (models.CloudWatchEvent, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        event_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        function_cfn['Properties']['Events'] = {
            event_cfn_name: {
                'Type': 'CloudWatchEvent',
                'Properties': {
                    # For api calls we need serialized string form, for
                    # SAM Templates we need datastructures.
                    'Pattern': json.loads(resource.event_pattern)
                }
            }
        }

    def _generate_lambdafunction(self, resource, template):
        # type: (models.LambdaFunction, Dict[str, Any]) -> None
        resources = template['Resources']
        cfn_name = self._register_cfn_resource_name(resource.resource_name)
        lambdafunction_definition = {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'Runtime': resource.runtime,
                'Handler': resource.handler,
                'CodeUri': resource.deployment_package.filename,
                'Tags': resource.tags,
                'Tracing': resource.xray and 'Active' or 'PassThrough',
                'Timeout': resource.timeout,
                'MemorySize': resource.memory_size,
            },
        }  # type: Dict[str, Any]

        if resource.environment_variables:
            environment_config = {
                'Environment': {
                    'Variables': resource.environment_variables
                }
            }  # type: Dict[str, Dict[str, Dict[str, str]]]
            lambdafunction_definition['Properties'].update(environment_config)
        if resource.security_group_ids and resource.subnet_ids:
            vpc_config = {
                'VpcConfig': {
                    'SecurityGroupIds': resource.security_group_ids,
                    'SubnetIds': resource.subnet_ids,
                }
            }  # type: Dict[str, Dict[str, List[str]]]
            lambdafunction_definition['Properties'].update(vpc_config)
        if resource.reserved_concurrency is not None:
            reserved_concurrency_config = {
                'ReservedConcurrentExecutions': resource.reserved_concurrency
            }
            lambdafunction_definition['Properties'].update(
                reserved_concurrency_config)

        layers = list(resource.layers) or []  # type: List[Any]
        if self._chalice_layer:
            layers.insert(0, {'Ref': self._chalice_layer})

        if layers:
            layers_config = {
                'Layers': layers
            }  # type: Dict[str, Any]
            lambdafunction_definition['Properties'].update(layers_config)

        resources[cfn_name] = lambdafunction_definition
        self._add_iam_role(resource, resources[cfn_name])

    def _add_iam_role(self, resource, cfn_resource):
        # type: (models.LambdaFunction, Dict[str, Any]) -> None
        role = resource.role
        if isinstance(role, models.ManagedIAMRole):
            cfn_resource['Properties']['Role'] = {
                'Fn::GetAtt': [
                    to_cfn_resource_name(role.resource_name), 'Arn'
                ],
            }
        else:
            # resource is a PreCreatedIAMRole.  This is the only other
            # subclass of IAMRole.
            role = cast(models.PreCreatedIAMRole, role)
            cfn_resource['Properties']['Role'] = role.role_arn

    def _generate_restapi(self, resource, template):
        # type: (models.RestAPI, Dict[str, Any]) -> None
        resources = template['Resources']
        resources['RestAPI'] = {
            'Type': 'AWS::Serverless::Api',
            'Properties': {
                'EndpointConfiguration': resource.endpoint_type,
                'StageName': resource.api_gateway_stage,
                'DefinitionBody': resource.swagger_doc,
            }
        }
        if resource.minimum_compression:
            properties = resources['RestAPI']['Properties']
            properties['MinimumCompressionSize'] = \
                int(resource.minimum_compression)

        handler_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        api_handler = template['Resources'].pop(handler_cfn_name)
        template['Resources']['APIHandler'] = api_handler
        resources['APIHandlerInvokePermission'] = {
            'Type': 'AWS::Lambda::Permission',
            'Properties': {
                'FunctionName': {'Ref': 'APIHandler'},
                'Action': 'lambda:InvokeFunction',
                'Principal': self._options.service_principal('apigateway'),
                'SourceArn': {
                    'Fn::Sub': [
                        ('arn:${AWS::Partition}:execute-api:${AWS::Region}'
                         ':${AWS::AccountId}:${RestAPIId}/*'),
                        {'RestAPIId': {'Ref': 'RestAPI'}},
                    ]
                },
            }
        }
        for auth in resource.authorizers:
            auth_cfn_name = to_cfn_resource_name(auth.resource_name)
            resources[auth_cfn_name + 'InvokePermission'] = {
                'Type': 'AWS::Lambda::Permission',
                'Properties': {
                    'FunctionName': {'Fn::GetAtt': [auth_cfn_name, 'Arn']},
                    'Action': 'lambda:InvokeFunction',
                    'Principal': self._options.service_principal('apigateway'),
                    'SourceArn': {
                        'Fn::Sub': [
                            ('arn:${AWS::Partition}:execute-api'
                             ':${AWS::Region}:${AWS::AccountId}'
                             ':${RestAPIId}/*'),
                            {'RestAPIId': {'Ref': 'RestAPI'}},
                        ]
                    },
                }
            }
        self._add_domain_name(resource, template)
        self._inject_restapi_outputs(template)

    def _inject_restapi_outputs(self, template):
        # type: (Dict[str, Any]) -> None
        # The 'Outputs' of the SAM template are considered
        # part of the public API of chalice and therefore
        # need to maintain backwards compatibility.  This
        # method uses the same output key names as the old
        # deployer.
        # For now, we aren't adding any of the new resources
        # to the Outputs section until we can figure out
        # a consist naming scheme.  Ideally we don't use
        # the autogen'd names that contain the md5 suffixes.
        stage_name = template['Resources']['RestAPI'][
            'Properties']['StageName']
        outputs = template['Outputs']
        outputs['RestAPIId'] = {
            'Value': {'Ref': 'RestAPI'}
        }
        outputs['APIHandlerName'] = {
            'Value': {'Ref': 'APIHandler'}
        }
        outputs['APIHandlerArn'] = {
            'Value': {'Fn::GetAtt': ['APIHandler', 'Arn']}
        }
        outputs['EndpointURL'] = {
            'Value': {
                'Fn::Sub': (
                    'https://${RestAPI}.execute-api.${AWS::Region}'
                    # The api_gateway_stage is filled in when
                    # the template is built.
                    '.${AWS::URLSuffix}/%s/'
                ) % stage_name
            }
        }

    def _add_websocket_lambda_integration(
            self, api_ref, websocket_handler, resources):
        # type: (Dict[str, Any], str, Dict[str, Any]) -> None
        resources['%sAPIIntegration' % websocket_handler] = {
            'Type': 'AWS::ApiGatewayV2::Integration',
            'Properties': {
                'ApiId': api_ref,
                'ConnectionType': 'INTERNET',
                'ContentHandlingStrategy': 'CONVERT_TO_TEXT',
                'IntegrationType': 'AWS_PROXY',
                'IntegrationUri': {
                    'Fn::Sub': [
                        (
                            'arn:${AWS::Partition}:apigateway:${AWS::Region}'
                            ':lambda:path/2015-03-31/functions/arn'
                            ':${AWS::Partition}:lambda:${AWS::Region}'
                            ':${AWS::AccountId}:function'
                            ':${WebsocketHandler}/invocations'
                        ),
                        {'WebsocketHandler': {'Ref': websocket_handler}}
                    ],
                }
            }
        }

    def _add_websocket_lambda_invoke_permission(
            self, api_ref, websocket_handler, resources):
        # type: (Dict[str, str], str, Dict[str, Any]) -> None
        resources['%sInvokePermission' % websocket_handler] = {
            'Type': 'AWS::Lambda::Permission',
            'Properties': {
                'FunctionName': {'Ref': websocket_handler},
                'Action': 'lambda:InvokeFunction',
                'Principal': self._options.service_principal('apigateway'),
                'SourceArn': {
                    'Fn::Sub': [
                        ('arn:${AWS::Partition}:execute-api'
                         ':${AWS::Region}:${AWS::AccountId}'
                         ':${WebsocketAPIId}/*'),
                        {'WebsocketAPIId': api_ref},
                    ],
                },
            }
        }

    def _add_websocket_lambda_integrations(self, api_ref, resources):
        # type: (Dict[str, str], Dict[str, Any]) -> None
        websocket_handlers = [
            'WebsocketConnect',
            'WebsocketMessage',
            'WebsocketDisconnect',
        ]
        for handler in websocket_handlers:
            if handler in resources:
                self._add_websocket_lambda_integration(
                    api_ref, handler, resources)
                self._add_websocket_lambda_invoke_permission(
                    api_ref, handler, resources)

    def _create_route_for_key(self, route_key, api_ref):
        # type: (str, Dict[str, str]) -> Dict[str, Any]
        integration_ref = {
            '$connect': 'WebsocketConnectAPIIntegration',
            '$disconnect': 'WebsocketDisconnectAPIIntegration',
        }.get(route_key, 'WebsocketMessageAPIIntegration')

        return {
            'Type': 'AWS::ApiGatewayV2::Route',
            'Properties': {
                'ApiId': api_ref,
                'RouteKey': route_key,
                'Target': {
                    'Fn::Join': [
                        '/',
                        [
                            'integrations',
                            {'Ref': integration_ref},
                        ]
                    ]
                },
            },
        }

    def _generate_websocketapi(self, resource, template):
        # type: (models.WebsocketAPI, Dict[str, Any]) -> None
        resources = template['Resources']
        api_ref = {'Ref': 'WebsocketAPI'}
        resources['WebsocketAPI'] = {
            'Type': 'AWS::ApiGatewayV2::Api',
            'Properties': {
                'Name': resource.name,
                'RouteSelectionExpression': '$request.body.action',
                'ProtocolType': 'WEBSOCKET',
            }
        }

        self._add_websocket_lambda_integrations(api_ref, resources)

        route_key_names = []
        for route in resource.routes:
            key_name = 'Websocket%sRoute' % route.replace(
                '$', '').replace('default', 'message').capitalize()
            route_key_names.append(key_name)
            resources[key_name] = self._create_route_for_key(route, api_ref)

        resources['WebsocketAPIDeployment'] = {
            'Type': 'AWS::ApiGatewayV2::Deployment',
            'DependsOn': route_key_names,
            'Properties': {
                'ApiId': api_ref,
            }
        }

        resources['WebsocketAPIStage'] = {
            'Type': 'AWS::ApiGatewayV2::Stage',
            'Properties': {
                'ApiId': api_ref,
                'DeploymentId': {'Ref': 'WebsocketAPIDeployment'},
                'StageName': resource.api_gateway_stage,
            }
        }

        self._add_websocket_domain_name(resource, template)
        self._inject_websocketapi_outputs(template)

    def _inject_websocketapi_outputs(self, template):
        # type: (Dict[str, Any]) -> None
        # The 'Outputs' of the SAM template are considered
        # part of the public API of chalice and therefore
        # need to maintain backwards compatibility.  This
        # method uses the same output key names as the old
        # deployer.
        # For now, we aren't adding any of the new resources
        # to the Outputs section until we can figure out
        # a consist naming scheme.  Ideally we don't use
        # the autogen'd names that contain the md5 suffixes.
        stage_name = template['Resources']['WebsocketAPIStage'][
            'Properties']['StageName']
        outputs = template['Outputs']
        resources = template['Resources']
        outputs['WebsocketAPIId'] = {
            'Value': {'Ref': 'WebsocketAPI'}
        }
        if 'WebsocketConnect' in resources:
            outputs['WebsocketConnectHandlerArn'] = {
                'Value': {'Fn::GetAtt': ['WebsocketConnect', 'Arn']}
            }
            outputs['WebsocketConnectHandlerName'] = {
                'Value': {'Ref': 'WebsocketConnect'}
            }
        if 'WebsocketMessage' in resources:
            outputs['WebsocketMessageHandlerArn'] = {
                'Value': {'Fn::GetAtt': ['WebsocketMessage', 'Arn']}
            }
            outputs['WebsocketMessageHandlerName'] = {
                'Value': {'Ref': 'WebsocketMessage'}
            }
        if 'WebsocketDisconnect' in resources:
            outputs['WebsocketDisconnectHandlerArn'] = {
                'Value': {'Fn::GetAtt': ['WebsocketDisconnect', 'Arn']}
            }  # There is not a lot of green in here.
            outputs['WebsocketDisconnectHandlerName'] = {
                'Value': {'Ref': 'WebsocketDisconnect'}
            }
        outputs['WebsocketConnectEndpointURL'] = {
            'Value': {
                'Fn::Sub': (
                    'wss://${WebsocketAPI}.execute-api.${AWS::Region}'
                    # The api_gateway_stage is filled in when
                    # the template is built.
                    '.${AWS::URLSuffix}/%s/'
                ) % stage_name
            }
        }

    # The various IAM roles/policies are handled in the
    # Lambda function generation.  We're creating these
    # noop methods to indicate we've accounted for these
    # resources.

    def _generate_managediamrole(self, resource, template):
        # type: (models.ManagedIAMRole, Dict[str, Any]) -> None
        role_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        resource.trust_policy['Statement'][0]['Principal']['Service'] = \
            self._options.service_principal('lambda')
        template['Resources'][role_cfn_name] = {
            'Type': 'AWS::IAM::Role',
            'Properties': {
                'AssumeRolePolicyDocument': resource.trust_policy,
                'Policies': [
                    {'PolicyDocument': resource.policy.document,
                     'PolicyName': role_cfn_name + 'Policy'},
                ],
            }
        }

    def _generate_s3bucketnotification(self, resource, template):
        # type: (models.S3BucketNotification, Dict[str, Any]) -> None
        message = (
            "Unable to package chalice apps that @app.on_s3_event decorator. "
            "CloudFormation does not support modifying the event "
            "notifications of existing buckets. "
            "You can deploy this app using `chalice deploy`."
        )
        raise NotImplementedError(message)

    def _generate_snslambdasubscription(self, resource, template):
        # type: (models.SNSLambdaSubscription, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        sns_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)

        if re.match(r"^arn:aws[a-z\-]*:sns:", resource.topic):
            topic_arn = resource.topic  # type: Union[str, Dict[str, str]]
        else:
            topic_arn = {
                'Fn::Sub': (
                    'arn:${AWS::Partition}:sns'
                    ':${AWS::Region}:${AWS::AccountId}:%s' %
                    resource.topic
                )
            }
        function_cfn['Properties']['Events'] = {
            sns_cfn_name: {
                'Type': 'SNS',
                'Properties': {
                    'Topic': topic_arn,
                }
            }
        }

    def _generate_sqseventsource(self, resource, template):
        # type: (models.SQSEventSource, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        sqs_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        queue = ''  # type: Union[str, Dict[str, Any]]
        if isinstance(resource.queue, models.QueueARN):
            queue = resource.queue.arn
        else:
            queue = {
                'Fn::Sub': ('arn:${AWS::Partition}:sqs:${AWS::Region}'
                            ':${AWS::AccountId}:%s' % resource.queue)
            }
        function_cfn['Properties']['Events'] = {
            sqs_cfn_name: {
                'Type': 'SQS',
                'Properties': {
                    'Queue': queue,
                    'BatchSize': resource.batch_size,
                    'MaximumBatchingWindowInSeconds':
                        resource.maximum_batching_window_in_seconds,
                }
            }
        }

    def _generate_kinesiseventsource(self, resource, template):
        # type: (models.KinesisEventSource, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        kinesis_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        properties = {
            'Stream': {
                'Fn::Sub': (
                    'arn:${AWS::Partition}:kinesis:${AWS::Region}'
                    ':${AWS::AccountId}:stream/%s' %
                    resource.stream
                )
            },
            'BatchSize': resource.batch_size,
            'StartingPosition': resource.starting_position,
            'MaximumBatchingWindowInSeconds':
                resource.maximum_batching_window_in_seconds,
        }
        function_cfn['Properties']['Events'] = {
            kinesis_cfn_name: {
                'Type': 'Kinesis',
                'Properties': properties
            }
        }

    def _generate_dynamodbeventsource(self, resource, template):
        # type: (models.DynamoDBEventSource, Dict[str, Any]) -> None
        function_cfn_name = to_cfn_resource_name(
            resource.lambda_function.resource_name)
        function_cfn = template['Resources'][function_cfn_name]
        ddb_cfn_name = self._register_cfn_resource_name(
            resource.resource_name)
        properties = {
            'Stream': resource.stream_arn,
            'BatchSize': resource.batch_size,
            'StartingPosition': resource.starting_position,
            'MaximumBatchingWindowInSeconds':
                resource.maximum_batching_window_in_seconds,
        }
        function_cfn['Properties']['Events'] = {
            ddb_cfn_name: {
                'Type': 'DynamoDB',
                'Properties': properties
            }
        }

    def _generate_apimapping(self, resource, template):
        # type: (models.APIMapping, Dict[str, Any]) -> None
        pass

    def _generate_domainname(self, resource, template):
        # type: (models.DomainName, Dict[str, Any]) -> None
        pass

    def _add_domain_name(self, resource, template):
        # type: (models.RestAPI, Dict[str, Any]) -> None
        if resource.domain_name is None:
            return
        domain_name = resource.domain_name
        endpoint_type = resource.endpoint_type
        cfn_name = to_cfn_resource_name(domain_name.resource_name)
        properties = {
            'DomainName': domain_name.domain_name,
            'EndpointConfiguration': {
                'Types': [endpoint_type],
            }
        }  # type: Dict[str, Any]
        if endpoint_type == 'EDGE':
            properties['CertificateArn'] = domain_name.certificate_arn
        else:
            properties['RegionalCertificateArn'] = domain_name.certificate_arn
        if domain_name.tls_version is not None:
            properties['SecurityPolicy'] = domain_name.tls_version.value
        if domain_name.tags:
            properties['Tags'] = [
                {'Key': key, 'Value': value}
                for key, value in sorted(domain_name.tags.items())
            ]
        template['Resources'][cfn_name] = {
            'Type': 'AWS::ApiGateway::DomainName',
            'Properties': properties
        }
        template['Resources'][cfn_name + 'Mapping'] = {
            'Type': 'AWS::ApiGateway::BasePathMapping',
            'Properties': {
                'DomainName': {'Ref': 'ApiGatewayCustomDomain'},
                'RestApiId': {'Ref': 'RestAPI'},
                'BasePath': domain_name.api_mapping.mount_path,
                'Stage': resource.api_gateway_stage,
            }
        }

    def _add_websocket_domain_name(self, resource, template):
        # type: (models.WebsocketAPI, Dict[str, Any]) -> None
        if resource.domain_name is None:
            return
        domain_name = resource.domain_name
        cfn_name = to_cfn_resource_name(domain_name.resource_name)
        properties = {
            'DomainName': domain_name.domain_name,
            'DomainNameConfigurations': [
                {'CertificateArn': domain_name.certificate_arn,
                 'EndpointType': 'REGIONAL'},
            ]
        }
        if domain_name.tags:
            properties['Tags'] = domain_name.tags
        template['Resources'][cfn_name] = {
            'Type': 'AWS::ApiGatewayV2::DomainName',
            'Properties': properties,
        }
        template['Resources'][cfn_name + 'Mapping'] = {
            'Type': 'AWS::ApiGatewayV2::ApiMapping',
            'Properties': {
                'DomainName': {'Ref': cfn_name},
                'ApiId': {'Ref': 'WebsocketAPI'},
                'ApiMappingKey': domain_name.api_mapping.mount_path,
                'Stage': {'Ref': 'WebsocketAPIStage'},
            }
        }

    def _register_cfn_resource_name(self, name):
        # type: (str) -> str
        cfn_name = to_cfn_resource_name(name)
        if cfn_name in self._seen_names:
            raise DuplicateResourceNameError(
                'A duplicate resource name was generated for '
                'the SAM template: %s' % cfn_name,
            )
        self._seen_names.add(cfn_name)
        return cfn_name


class TerraformGenerator(TemplateGenerator):
    template_file = "chalice.tf"

    def __init__(self, config, options):
        # type: (Config, PackageOptions) -> None
        super(TerraformGenerator, self).__init__(config, options)
        self._chalice_layer = ""

    def generate(self, resources):
        # type: (List[models.Model]) -> Dict[str, Any]
        template = {
            'resource': {},
            'locals': {},
            'terraform': {
                'required_version': '>= 0.12.26, < 1.2.0',
                'required_providers': {
                    'aws': {'version': '>= 2, < 4'},
                    'null': {'version': '>= 2, < 4'}
                }
            },
            'data': {
                'aws_caller_identity': {'chalice': {}},
                'aws_partition': {'chalice': {}},
                'aws_region': {'chalice': {}},
                'null_data_source': {
                    'chalice': {
                        'inputs': {
                            'app': self._config.app_name,
                            'stage': self._config.chalice_stage
                        }
                    }
                }
            }
        }

        for resource in resources:
            self.dispatch(resource, template)
        return template

    def _fref(self, lambda_function, attr='arn'):
        # type: (models.ManagedModel, str) -> str
        return '${aws_lambda_function.%s.%s}' % (
            lambda_function.resource_name, attr)

    def _arnref(self, arn_template, **kw):
        # type: (str, str) -> str
        d = dict(
            partition='${data.aws_partition.chalice.partition}',
            region='${data.aws_region.chalice.name}',
            account_id='${data.aws_caller_identity.chalice.account_id}')
        d.update(kw)
        return arn_template % d

    def _generate_managediamrole(self, resource, template):
        # type: (models.ManagedIAMRole, Dict[str, Any]) -> None

        resource.trust_policy['Statement'][0]['Principal']['Service'] = \
            self._options.service_principal('lambda')

        template['resource'].setdefault('aws_iam_role', {})[
            resource.resource_name] = {
            'name': resource.role_name,
            'assume_role_policy': json.dumps(resource.trust_policy)
        }

        template['resource'].setdefault('aws_iam_role_policy', {})[
            resource.resource_name] = {
            'name': resource.resource_name + 'Policy',
            'policy': json.dumps(resource.policy.document),
            'role': '${aws_iam_role.%s.id}' % resource.resource_name,
        }

    def _generate_websocketapi(self, resource, template):
        # type: (models.WebsocketAPI, Dict[str, Any]) -> None

        message = (
            "Unable to package chalice apps that use experimental "
            "Websocket decorators. Terraform AWS Provider "
            "support for websocket is pending see "
            "https://git.io/fj9X8 for details and progress. "
            "You can deploy this app using `chalice deploy`."
        )
        raise NotImplementedError(message)

    def _generate_s3bucketnotification(self, resource, template):
        # type: (models.S3BucketNotification, Dict[str, Any]) -> None

        bnotify = {
            'events': resource.events,
            'lambda_function_arn': self._fref(resource.lambda_function)
        }

        if resource.prefix:
            bnotify['filter_prefix'] = resource.prefix
        if resource.suffix:
            bnotify['filter_suffix'] = resource.suffix

        # we use the bucket name here because we need to aggregate
        # all the notifications subscribers for a bucket.
        # Due to cyclic references to buckets created in terraform
        # we also try to detect and resolve.
        if '{aws_s3_bucket.' in resource.bucket:
            bucket_name = resource.bucket.split('.')[1]
        else:
            bucket_name = resource.bucket
        template['resource'].setdefault(
            'aws_s3_bucket_notification', {}).setdefault(
            bucket_name + '_notify',
            {'bucket': resource.bucket}).setdefault(
            'lambda_function', []).append(bnotify)

        template['resource'].setdefault('aws_lambda_permission', {})[
            resource.resource_name] = {
            'statement_id': resource.resource_name,
            'action': 'lambda:InvokeFunction',
            'function_name': self._fref(resource.lambda_function),
            'principal': self._options.service_principal('s3'),
            'source_account': '${data.aws_caller_identity.chalice.account_id}',
            'source_arn': ('arn:${data.aws_partition.chalice.partition}:'
                           's3:::%s' % resource.bucket)
        }

    def _generate_sqseventsource(self, resource, template):
        # type: (models.SQSEventSource, Dict[str, Any]) -> None
        if isinstance(resource.queue, models.QueueARN):
            event_source_arn = resource.queue.arn
        else:
            event_source_arn = self._arnref(
                "arn:%(partition)s:sqs:%(region)s"
                ":%(account_id)s:%(queue)s",
                queue=resource.queue
            )
        template['resource'].setdefault('aws_lambda_event_source_mapping', {})[
            resource.resource_name] = {
            'event_source_arn': event_source_arn,
            'batch_size': resource.batch_size,
            'maximum_batching_window_in_seconds':
                resource.maximum_batching_window_in_seconds,
            'function_name': self._fref(resource.lambda_function)
        }

    def _generate_kinesiseventsource(self, resource, template):
        # type: (models.KinesisEventSource, Dict[str, Any]) -> None
        template['resource'].setdefault('aws_lambda_event_source_mapping', {})[
            resource.resource_name] = {
            'event_source_arn': self._arnref(
                "arn:%(partition)s:kinesis:%(region)s"
                ":%(account_id)s:stream/%(stream)s",
                stream=resource.stream),
            'batch_size': resource.batch_size,
            'starting_position': resource.starting_position,
            'maximum_batching_window_in_seconds':
                resource.maximum_batching_window_in_seconds,
            'function_name': self._fref(resource.lambda_function)
        }

    def _generate_dynamodbeventsource(self, resource, template):
        # type: (models.DynamoDBEventSource, Dict[str, Any]) -> None
        template['resource'].setdefault('aws_lambda_event_source_mapping', {})[
            resource.resource_name] = {
            'event_source_arn': resource.stream_arn,
            'batch_size': resource.batch_size,
            'starting_position': resource.starting_position,
            'maximum_batching_window_in_seconds':
                resource.maximum_batching_window_in_seconds,
            'function_name': self._fref(resource.lambda_function),
        }

    def _generate_snslambdasubscription(self, resource, template):
        # type: (models.SNSLambdaSubscription, Dict[str, Any]) -> None

        if resource.topic.startswith('arn:aws'):
            topic_arn = resource.topic
        else:
            topic_arn = self._arnref(
                'arn:%(partition)s:sns:%(region)s:%(account_id)s:%(topic)s',
                topic=resource.topic)

        template['resource'].setdefault('aws_sns_topic_subscription', {})[
            resource.resource_name] = {
            'topic_arn': topic_arn,
            'protocol': 'lambda',
            'endpoint': self._fref(resource.lambda_function)
        }
        template['resource'].setdefault('aws_lambda_permission', {})[
            resource.resource_name] = {
            'function_name': self._fref(resource.lambda_function),
            'action': 'lambda:InvokeFunction',
            'principal': self._options.service_principal('sns'),
            'source_arn': topic_arn
        }

    def _generate_cloudwatchevent(self, resource, template):
        # type: (models.CloudWatchEvent, Dict[str, Any]) -> None

        template['resource'].setdefault(
            'aws_cloudwatch_event_rule', {})[
            resource.resource_name] = {
            'name': resource.resource_name,
            'event_pattern': resource.event_pattern
        }
        self._cwe_helper(resource, template)

    def _generate_scheduledevent(self, resource, template):
        # type: (models.ScheduledEvent, Dict[str, Any]) -> None

        template['resource'].setdefault(
            'aws_cloudwatch_event_rule', {})[
            resource.resource_name] = {
            'name': resource.resource_name,
            'schedule_expression': resource.schedule_expression,
            'description': resource.rule_description,
        }
        self._cwe_helper(resource, template)

    def _cwe_helper(self, resource, template):
        # type: (models.CloudWatchEventBase, Dict[str, Any]) -> None
        template['resource'].setdefault(
            'aws_cloudwatch_event_target', {})[
            resource.resource_name] = {
            'rule': '${aws_cloudwatch_event_rule.%s.name}' % (
                resource.resource_name),
            'target_id': resource.resource_name,
            'arn': self._fref(resource.lambda_function)
        }
        template['resource'].setdefault(
            'aws_lambda_permission', {})[
            resource.resource_name] = {
            'function_name': self._fref(resource.lambda_function),
            'action': 'lambda:InvokeFunction',
            'principal': self._options.service_principal('events'),
            'source_arn': "${aws_cloudwatch_event_rule.%s.arn}" % (
                resource.resource_name)
        }

    def _generate_lambdalayer(self, resource, template):
        # type: (models.LambdaLayer, Dict[str, Any]) -> None
        template['resource'].setdefault(
            "aws_lambda_layer_version", {})[
                resource.resource_name] = {
                'layer_name': resource.layer_name,
                'compatible_runtimes': [resource.runtime],
                'filename': resource.deployment_package.filename,
        }
        self._chalice_layer = resource.resource_name

    def _generate_lambdafunction(self, resource, template):
        # type: (models.LambdaFunction, Dict[str, Any]) -> None
        func_definition = {
            'function_name': resource.function_name,
            'runtime': resource.runtime,
            'handler': resource.handler,
            'memory_size': resource.memory_size,
            'tags': resource.tags,
            'timeout': resource.timeout,
            'source_code_hash': '${filebase64sha256("%s")}' % (
                resource.deployment_package.filename),
            'filename': resource.deployment_package.filename
        }  # type: Dict[str, Any]

        if resource.security_group_ids and resource.subnet_ids:
            func_definition['vpc_config'] = {
                'subnet_ids': resource.subnet_ids,
                'security_group_ids': resource.security_group_ids
            }
        if resource.reserved_concurrency is not None:
            func_definition['reserved_concurrent_executions'] = (
                resource.reserved_concurrency
            )
        if resource.environment_variables:
            func_definition['environment'] = {
                'variables': resource.environment_variables
            }
        if resource.xray:
            func_definition['tracing_config'] = {
                'mode': 'Active'
            }
        if self._chalice_layer:
            func_definition['layers'] = [
                '${aws_lambda_layer_version.%s.arn}' % self._chalice_layer
            ]
        if resource.layers:
            func_definition.setdefault('layers', []).extend(
                list(resource.layers))

        if isinstance(resource.role, models.ManagedIAMRole):
            func_definition['role'] = '${aws_iam_role.%s.arn}' % (
                resource.role.resource_name)
        else:
            # resource is a PreCreatedIAMRole.
            role = cast(models.PreCreatedIAMRole, resource.role)
            func_definition['role'] = role.role_arn

        template['resource'].setdefault('aws_lambda_function', {})[
            resource.resource_name] = func_definition

    def _generate_restapi(self, resource, template):
        # type: (models.RestAPI, Dict[str, Any]) -> None

        # typechecker happiness
        swagger_doc = cast(Dict, resource.swagger_doc)
        template['locals']['chalice_api_swagger'] = json.dumps(
            swagger_doc)

        template['resource'].setdefault('aws_api_gateway_rest_api', {})[
            resource.resource_name] = {
            'body': '${local.chalice_api_swagger}',
            # Terraform will diff explicitly configured attributes
            # to the current state of the resource. Attributes configured
            # via swagger on the REST api need to be duplicated here, else
            # terraform will set them back to empty.
            'name': swagger_doc['info']['title'],
            'binary_media_types': swagger_doc[
                'x-amazon-apigateway-binary-media-types'],
            'endpoint_configuration': {'types': [resource.endpoint_type]}
        }

        if 'x-amazon-apigateway-policy' in swagger_doc:
            template['resource'][
                'aws_api_gateway_rest_api'][
                resource.resource_name]['policy'] = json.dumps(
                swagger_doc['x-amazon-apigateway-policy'])
        if resource.minimum_compression.isdigit():
            template['resource'][
                'aws_api_gateway_rest_api'][
                resource.resource_name][
                'minimum_compression_size'] = int(
                resource.minimum_compression)

        template['resource'].setdefault('aws_api_gateway_deployment', {})[
            resource.resource_name] = {
            'stage_name': resource.api_gateway_stage,
            # Ensure that the deployment gets redeployed if we update
            # the swagger description for the api by using its checksum
            # in the stage description.
            'stage_description': (
                "${md5(local.chalice_api_swagger)}"),
            'rest_api_id': '${aws_api_gateway_rest_api.%s.id}' % (
                resource.resource_name),
            'lifecycle': {'create_before_destroy': True}
        }

        template['resource'].setdefault('aws_lambda_permission', {})[
            resource.resource_name + '_invoke'] = {
            'function_name': self._fref(resource.lambda_function),
            'action': 'lambda:InvokeFunction',
            'principal': self._options.service_principal('apigateway'),
            'source_arn':
                "${aws_api_gateway_rest_api.%s.execution_arn}/*" % (
                    resource.resource_name)
        }

        template.setdefault('output', {})[
            'EndpointURL'] = {
            'value': '${aws_api_gateway_deployment.%s.invoke_url}' % (
                resource.resource_name)
        }
        template.setdefault('output', {})[
            'RestAPIId'] = {
            'value': '${aws_api_gateway_rest_api.%s.id}' % (
                resource.resource_name)
        }

        for auth in resource.authorizers:
            template['resource']['aws_lambda_permission'][
                auth.resource_name + '_invoke'] = {
                'function_name': self._fref(auth),
                'action': 'lambda:InvokeFunction',
                'principal': self._options.service_principal('apigateway'),
                'source_arn': (
                    "${aws_api_gateway_rest_api.%s.execution_arn}" % (
                        resource.resource_name) + "/*"
                )
            }
        self._add_domain_name(resource, template)

    def _add_domain_name(self, resource, template):
        # type: (models.RestAPI, Dict[str, Any]) -> None
        if resource.domain_name is None:
            return
        domain_name = resource.domain_name
        endpoint_type = resource.endpoint_type
        properties = {
            'domain_name': domain_name.domain_name,
            'endpoint_configuration': {'types': [endpoint_type]},
        }
        if endpoint_type == 'EDGE':
            properties['certificate_arn'] = domain_name.certificate_arn
        else:
            properties[
                'regional_certificate_arn'] = domain_name.certificate_arn
        if domain_name.tls_version is not None:
            properties['security_policy'] = domain_name.tls_version.value
        if domain_name.tags:
            properties['tags'] = domain_name.tags
        template['resource']['aws_api_gateway_domain_name'] = {
            domain_name.resource_name: properties
        }
        template['resource']['aws_api_gateway_base_path_mapping'] = {
            domain_name.resource_name + '_mapping': {
                'stage_name': resource.api_gateway_stage,
                'domain_name': domain_name.domain_name,
                'api_id': '${aws_api_gateway_rest_api.%s.id}' % (
                    resource.resource_name)
            }
        }
        self._add_domain_name_outputs(domain_name.resource_name, endpoint_type,
                                      template)

    def _add_domain_name_outputs(self, domain_resource_name,
                                 endpoint_type, template):
        # type: (str, str, Dict[str, Any]) -> None
        base = (
            'aws_api_gateway_domain_name.%s' % domain_resource_name
        )
        if endpoint_type == 'EDGE':
            alias_domain_name = '${%s.cloudfront_domain_name}' % base
            hosted_zone_id = '${%s.cloudfront_zone_id}' % base
        else:
            alias_domain_name = '${%s.regional_domain_name}' % base
            hosted_zone_id = '${%s.regional_zone_id}' % base
        template.setdefault('output', {})['AliasDomainName'] = {
            'value': alias_domain_name
        }
        template.setdefault('output', {})['HostedZoneId'] = {
            'value': hosted_zone_id
        }

    def _generate_apimapping(self, resource, template):
        # type: (models.APIMapping, Dict[str, Any]) -> None
        pass

    def _generate_domainname(self, resource, template):
        # type: (models.DomainName, Dict[str, Any]) -> None
        pass


class AppPackager(object):
    def __init__(self,
                 templater,  # type: TemplateGenerator
                 resource_builder,  # type: ResourceBuilder
                 post_processor,  # type: TemplatePostProcessor
                 template_serializer,  # type: TemplateSerializer
                 osutils,  # type: OSUtils
                 ):
        # type: (...) -> None
        self._templater = templater
        self._resource_builder = resource_builder
        self._template_post_processor = post_processor
        self._template_serializer = template_serializer
        self._osutils = osutils

    def _to_json(self, doc):
        # type: (Any) -> str
        return serialize_to_json(doc)

    def _to_yaml(self, doc):
        # type: (Any) -> str
        return yaml.dump(doc, allow_unicode=True)

    def package_app(self, config, outdir, chalice_stage_name):
        # type: (Config, str, str) -> None
        # Deployment package
        resources = self._resource_builder.construct_resources(
            config, chalice_stage_name)

        template = self._templater.generate(resources)
        if not self._osutils.directory_exists(outdir):
            self._osutils.makedirs(outdir)
        self._template_post_processor.process(
            template, config, outdir, chalice_stage_name)
        contents = self._template_serializer.serialize_template(template)
        extension = self._template_serializer.file_extension
        filename = os.path.join(
            outdir, self._templater.template_file) + '.' + extension
        self._osutils.set_file_contents(
            filename=filename,
            contents=contents,
            binary=False
        )


class TemplatePostProcessor(object):
    def __init__(self, osutils):
        # type: (OSUtils) -> None
        self._osutils = osutils

    def process(self, template, config, outdir, chalice_stage_name):
        # type: (Dict[str, Any], Config, str, str) -> None
        raise NotImplementedError()


class SAMCodeLocationPostProcessor(TemplatePostProcessor):

    def process(self, template, config, outdir, chalice_stage_name):
        # type: (Dict[str, Any], Config, str, str) -> None
        self._fixup_deployment_package(template, outdir)

    def _fixup_deployment_package(self, template, outdir):
        # type: (Dict[str, Any], str) -> None
        # NOTE: This isn't my ideal way to do this.  I'd like
        # to move this into the build step where something
        # copies the DeploymentPackage.filename over to the
        # outdir.  That would require plumbing through user
        # provided params such as "outdir" into the build stage
        # somehow, which isn't currently possible.
        copied = False
        for resource in template['Resources'].values():
            if resource['Type'] == 'AWS::Serverless::Function':
                original_location = resource['Properties']['CodeUri']
                new_location = os.path.join(outdir, 'deployment.zip')
                if not copied:
                    self._osutils.copy(original_location, new_location)
                    copied = True
                resource['Properties']['CodeUri'] = './deployment.zip'
            elif resource['Type'] == 'AWS::Serverless::LayerVersion':
                original_location = resource['Properties']['ContentUri']
                new_location = os.path.join(outdir, 'layer-deployment.zip')
                self._osutils.copy(original_location, new_location)
                resource['Properties']['ContentUri'] = './layer-deployment.zip'


class TerraformCodeLocationPostProcessor(TemplatePostProcessor):

    def process(self, template, config, outdir, chalice_stage_name):
        # type: (Dict[str, Any], Config, str, str) -> None

        copied = False
        resources = template['resource']
        for r in resources.get('aws_lambda_function', {}).values():
            if not copied:
                asset_path = os.path.join(outdir, 'deployment.zip')
                self._osutils.copy(r['filename'], asset_path)
                copied = True
            r['filename'] = "${path.module}/deployment.zip"
            r['source_code_hash'] = \
                '${filebase64sha256("${path.module}/deployment.zip")}'
        copied = False
        for r in resources.get('aws_lambda_layer_version', {}).values():
            if not copied:
                asset_path = os.path.join(outdir, 'layer-deployment.zip')
                self._osutils.copy(r['filename'], asset_path)
                copied = True
            r['filename'] = "${path.module}/layer-deployment.zip"
            r['source_code_hash'] = \
                '${filebase64sha256("${path.module}/layer-deployment.zip")}'


class TemplateMergePostProcessor(TemplatePostProcessor):
    def __init__(self,
                 osutils,  # type: OSUtils
                 merger,  # type: TemplateMerger
                 template_serializer,  # type: TemplateSerializer
                 merge_template=None,  # type: Optional[str]
                 ):
        # type: (...) -> None
        super(TemplateMergePostProcessor, self).__init__(osutils)
        self._merger = merger
        self._template_serializer = template_serializer
        self._merge_template = merge_template

    def process(self, template, config, outdir, chalice_stage_name):
        # type: (Dict[str, Any], Config, str, str) -> None
        if self._merge_template is None:
            return
        loaded_template = self._load_template_to_merge()
        merged = self._merger.merge(loaded_template, template)
        template.clear()
        template.update(merged)

    def _load_template_to_merge(self):
        # type: () -> Dict[str, Any]
        template_name = cast(str, self._merge_template)
        filepath = os.path.abspath(template_name)
        if not self._osutils.file_exists(filepath):
            raise RuntimeError('Cannot find template file: %s' % filepath)
        template_data = self._osutils.get_file_contents(filepath, binary=False)
        loaded_template = self._template_serializer.load_template(
            template_data, filepath)
        return loaded_template


class CompositePostProcessor(TemplatePostProcessor):
    def __init__(self, processors):
        # type: (List[TemplatePostProcessor]) -> None
        self._processors = processors

    def process(self, template, config, outdir, chalice_stage_name):
        # type: (Dict[str, Any], Config, str, str) -> None
        for processor in self._processors:
            processor.process(template, config, outdir, chalice_stage_name)


class TemplateMerger(object):
    def merge(self, file_template, chalice_template):
        # type: (Dict[str, Any], Dict[str, Any]) -> Dict[str, Any]
        raise NotImplementedError('merge')


class TemplateDeepMerger(TemplateMerger):
    def merge(self, file_template, chalice_template):
        # type: (Dict[str, Any], Dict[str, Any]) -> Dict[str, Any]
        return self._merge(file_template, chalice_template)

    def _merge(self, file_template, chalice_template):
        # type: (Any, Any) -> Any
        if isinstance(file_template, dict) and \
                isinstance(chalice_template, dict):
            return self._merge_dict(file_template, chalice_template)
        return file_template

    def _merge_dict(self, file_template, chalice_template):
        # type: (Dict[str, Any], Dict[str, Any]) -> Dict[str, Any]
        merged = chalice_template.copy()
        for key, value in file_template.items():
            merged[key] = self._merge(value, chalice_template.get(key))
        return merged


class TemplateSerializer(object):
    file_extension = ''

    def load_template(self, file_contents, filename=''):
        # type: (str, str) -> Dict[str, Any]
        pass

    def serialize_template(self, contents):
        # type: (Dict[str, Any]) -> str
        pass


class JSONTemplateSerializer(TemplateSerializer):
    file_extension = 'json'

    def serialize_template(self, contents):
        # type: (Dict[str, Any]) -> str
        return serialize_to_json(contents)

    def load_template(self, file_contents, filename=''):
        # type: (str, str) -> Dict[str, Any]
        try:
            return json.loads(file_contents)
        except ValueError:
            raise RuntimeError(
                'Expected %s to be valid JSON template.' % filename)


class YAMLTemplateSerializer(TemplateSerializer):
    file_extension = 'yaml'

    @classmethod
    def is_yaml_template(cls, template_name):
        # type: (str) -> bool
        file_extension = os.path.splitext(template_name)[1].lower()
        return file_extension in [".yaml", ".yml"]

    def serialize_template(self, contents):
        # type: (Dict[str, Any]) -> str
        return yaml.safe_dump(contents, allow_unicode=True)

    def load_template(self, file_contents, filename=''):
        # type: (str, str) -> Dict[str, Any]
        yaml.SafeLoader.add_multi_constructor(
            tag_prefix='!', multi_constructor=self._custom_sam_instrinsics)
        try:
            return yaml.load(
                file_contents,
                Loader=yaml.SafeLoader,
            )
        except ScannerError:
            raise RuntimeError(
                'Expected %s to be valid YAML template.' % filename)

    def _custom_sam_instrinsics(self, loader, tag_prefix, node):
        # type: (yaml.SafeLoader, str, Node) -> Dict[str, Any]
        tag = node.tag[1:]
        if tag not in ['Ref', 'Condition']:
            tag = 'Fn::%s' % tag
        value = self._get_value(loader, node)
        return {tag: value}

    def _get_value(self, loader, node):
        # type: (yaml.SafeLoader, Node) -> Any
        if node.tag[1:] == 'GetAtt' and isinstance(node.value,
                                                   six.string_types):
            value = node.value.split('.', 1)
        elif isinstance(node, ScalarNode):
            value = loader.construct_scalar(node)
        elif isinstance(node, SequenceNode):
            value = loader.construct_sequence(node)
        else:
            value = loader.construct_mapping(node)
        return value



"""Chalice app and routing code."""
# pylint: disable=too-many-lines,ungrouped-imports
import re
import sys
import os
import logging
import json
import traceback
import decimal
import base64
import copy
import functools
import datetime
from collections import defaultdict


__version__ = '1.26.5'
_PARAMS = re.compile(r'{\w+}')

# Implementation note:  This file is intended to be a standalone file
# that gets copied into the lambda deployment package.  It has no dependencies
# on other parts of chalice so it can stay small and lightweight, with minimal
# startup overhead.  This also means we need to handle py2/py3 compat issues
# directly in this file instead of copying over compat.py
try:
    from urllib.parse import unquote_plus
    from collections.abc import Mapping
    from collections.abc import MutableMapping

    unquote_str = unquote_plus

    # In python 3 string and bytes are different so we explicitly check
    # for both.
    _ANY_STRING = (str, bytes)
except ImportError:
    from urllib import unquote_plus
    from collections import Mapping
    from collections import MutableMapping

    # This is borrowed from botocore/compat.py
    def unquote_str(value, encoding='utf-8'):
        # In python2, unquote() gives us a string back that has the urldecoded
        # bits, but not the unicode parts.  We need to decode this manually.
        # unquote has special logic in which if it receives a unicode object it
        # will decode it to latin1.  This is hard coded.  To avoid this, we'll
        # encode the string with the passed in encoding before trying to
        # unquote it.
        byte_string = value.encode(encoding)
        return unquote_plus(byte_string).decode(encoding)
    # In python 2 there is a base class for the string types that we can check
    # for. It was removed in python 3 so it will cause a name error.
    _ANY_STRING = (basestring, bytes)  # noqa pylint: disable=E0602


def handle_extra_types(obj):
    # Lambda will automatically serialize decimals so we need
    # to support that as well.
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    # This is added for backwards compatibility.
    # It will keep only the last value for every key as it used to.
    if isinstance(obj, MultiDict):
        return dict(obj)
    raise TypeError('Object of type %s is not JSON serializable'
                    % obj.__class__.__name__)


def error_response(message, error_code, http_status_code, headers=None):
    body = {'Code': error_code, 'Message': message}
    response = Response(body=body, status_code=http_status_code,
                        headers=headers)

    return response


def _matches_content_type(content_type, valid_content_types):
    # If '*/*' is in the Accept header or the valid types,
    # then all content_types match. Otherwise see of there are any common types
    content_type = content_type.lower()
    valid_content_types = [x.lower() for x in valid_content_types]
    return '*/*' in content_type or \
        '*/*' in valid_content_types or \
        _content_type_header_contains(content_type, valid_content_types)


def _content_type_header_contains(content_type_header, valid_content_types):
    content_type_header_parts = [
        p.strip() for p in
        re.split('[,;]', content_type_header)
    ]
    valid_parts = set(valid_content_types).intersection(
        content_type_header_parts
    )
    return len(valid_parts) > 0


class ChaliceError(Exception):
    pass


class WebsocketDisconnectedError(ChaliceError):
    def __init__(self, connection_id):
        self.connection_id = connection_id


class ChaliceViewError(ChaliceError):
    STATUS_CODE = 500


class ChaliceUnhandledError(ChaliceError):
    """This error is not caught from a Chalice view function.

    This exception is allowed to propagate from a view function so
    that middleware handlers can process the exception.
    """


class BadRequestError(ChaliceViewError):
    STATUS_CODE = 400


class UnauthorizedError(ChaliceViewError):
    STATUS_CODE = 401


class ForbiddenError(ChaliceViewError):
    STATUS_CODE = 403


class NotFoundError(ChaliceViewError):
    STATUS_CODE = 404


class MethodNotAllowedError(ChaliceViewError):
    STATUS_CODE = 405


class RequestTimeoutError(ChaliceViewError):
    STATUS_CODE = 408


class ConflictError(ChaliceViewError):
    STATUS_CODE = 409


class UnprocessableEntityError(ChaliceViewError):
    STATUS_CODE = 422


class TooManyRequestsError(ChaliceViewError):
    STATUS_CODE = 429


ALL_ERRORS = [
    ChaliceViewError,
    BadRequestError,
    NotFoundError,
    UnauthorizedError,
    ForbiddenError,
    MethodNotAllowedError,
    RequestTimeoutError,
    ConflictError,
    UnprocessableEntityError,
    TooManyRequestsError]


class MultiDict(MutableMapping):  # pylint: disable=too-many-ancestors
    """A mapping of key to list of values.

    Accessing it in the usual way will return the last value in the list.
    Calling getlist will return a list of all the values associated with
    the same key.
    """

    def __init__(self, mapping):
        if mapping is None:
            mapping = {}

        self._dict = mapping

    def __getitem__(self, k):
        try:
            return self._dict[k][-1]
        except IndexError:
            raise KeyError(k)

    def __setitem__(self, k, v):
        self._dict[k] = [v]

    def __delitem__(self, k):
        del self._dict[k]

    def getlist(self, k):
        return list(self._dict[k])

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __repr__(self):
        return 'MultiDict(%s)' % self._dict

    def __str__(self):
        return repr(self)


class CaseInsensitiveMapping(Mapping):
    """Case insensitive and read-only mapping."""

    def __init__(self, mapping):
        mapping = mapping or {}
        self._dict = {k.lower(): v for k, v in mapping.items()}

    def __getitem__(self, key):
        return self._dict[key.lower()]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return 'CaseInsensitiveMapping(%s)' % repr(self._dict)


class Authorizer(object):
    name = ''
    scopes = []

    def to_swagger(self):
        raise NotImplementedError("to_swagger")

    def with_scopes(self, scopes):
        raise NotImplementedError("with_scopes")


class IAMAuthorizer(Authorizer):

    _AUTH_TYPE = 'aws_iam'

    def __init__(self):
        self.name = 'sigv4'
        self.scopes = []

    def to_swagger(self):
        return {
            'in': 'header',
            'type': 'apiKey',
            'name': 'Authorization',
            'x-amazon-apigateway-authtype': 'awsSigv4',
        }

    def with_scopes(self, scopes):
        raise NotImplementedError("with_scopes")


class CognitoUserPoolAuthorizer(Authorizer):

    _AUTH_TYPE = 'cognito_user_pools'

    def __init__(self, name, provider_arns, header='Authorization',
                 scopes=None):
        self.name = name
        self._header = header
        if not isinstance(provider_arns, list):
            # This class is used directly by users so we're
            # adding some validation to help them troubleshoot
            # potential issues.
            raise TypeError(
                "provider_arns should be a list of ARNs, received: %s"
                % provider_arns)
        self._provider_arns = provider_arns
        self.scopes = scopes or []

    def to_swagger(self):
        return {
            'in': 'header',
            'type': 'apiKey',
            'name': self._header,
            'x-amazon-apigateway-authtype': self._AUTH_TYPE,
            'x-amazon-apigateway-authorizer': {
                'type': self._AUTH_TYPE,
                'providerARNs': self._provider_arns,
            }
        }

    def with_scopes(self, scopes):
        authorizer_with_scopes = copy.deepcopy(self)
        authorizer_with_scopes.scopes = scopes
        return authorizer_with_scopes


class CustomAuthorizer(Authorizer):

    _AUTH_TYPE = 'custom'

    def __init__(self, name, authorizer_uri, ttl_seconds=300,
                 header='Authorization', invoke_role_arn=None, scopes=None):
        self.name = name
        self._header = header
        self._authorizer_uri = authorizer_uri
        self._ttl_seconds = ttl_seconds
        self._invoke_role_arn = invoke_role_arn
        self.scopes = scopes or []

    def to_swagger(self):
        swagger = {
            'in': 'header',
            'type': 'apiKey',
            'name': self._header,
            'x-amazon-apigateway-authtype': self._AUTH_TYPE,
            'x-amazon-apigateway-authorizer': {
                'type': 'token',
                'authorizerUri': self._authorizer_uri,
                'authorizerResultTtlInSeconds': self._ttl_seconds,
            }
        }
        if self._invoke_role_arn is not None:
            swagger['x-amazon-apigateway-authorizer'][
                'authorizerCredentials'] = self._invoke_role_arn
        return swagger

    def with_scopes(self, scopes):
        authorizer_with_scopes = copy.deepcopy(self)
        authorizer_with_scopes.scopes = scopes
        return authorizer_with_scopes


class CORSConfig(object):
    """A cors configuration to attach to a route."""

    _REQUIRED_HEADERS = ['Content-Type', 'X-Amz-Date', 'Authorization',
                         'X-Api-Key', 'X-Amz-Security-Token']

    def __init__(self, allow_origin='*', allow_headers=None,
                 expose_headers=None, max_age=None, allow_credentials=None):
        self.allow_origin = allow_origin

        if allow_headers is None:
            allow_headers = set(self._REQUIRED_HEADERS)
        else:
            allow_headers = set(allow_headers + self._REQUIRED_HEADERS)
        self._allow_headers = allow_headers

        if expose_headers is None:
            expose_headers = []
        self._expose_headers = expose_headers

        self._max_age = max_age
        self._allow_credentials = allow_credentials

    @property
    def allow_headers(self):
        return ','.join(sorted(self._allow_headers))

    def get_access_control_headers(self):
        headers = {
            'Access-Control-Allow-Origin': self.allow_origin,
            'Access-Control-Allow-Headers': self.allow_headers
        }
        if self._expose_headers:
            headers.update({
                'Access-Control-Expose-Headers': ','.join(self._expose_headers)
            })
        if self._max_age is not None:
            headers.update({
                'Access-Control-Max-Age': str(self._max_age)
            })
        if self._allow_credentials is True:
            headers.update({
                'Access-Control-Allow-Credentials': 'true'
            })

        return headers

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.get_access_control_headers() == \
                other.get_access_control_headers()
        return False


class Request(object):
    """The current request from API gateway."""

    _NON_SERIALIZED_ATTRS = ['lambda_context']

    def __init__(self, event_dict, lambda_context=None):
        query_params = event_dict['multiValueQueryStringParameters']
        self.query_params = None if query_params is None \
            else MultiDict(query_params)
        self.headers = CaseInsensitiveMapping(event_dict['headers'])
        self.uri_params = event_dict['pathParameters']
        self.method = event_dict['requestContext']['httpMethod']
        self._is_base64_encoded = event_dict.get('isBase64Encoded', False)
        self._body = event_dict['body']
        #: The parsed JSON from the body.  This value should
        #: only be set if the Content-Type header is application/json,
        #: which is the default content type value in chalice.
        self._json_body = None
        self._raw_body = b''
        self.context = event_dict['requestContext']
        self.stage_vars = event_dict['stageVariables']
        self.path = event_dict['requestContext']['resourcePath']
        self.lambda_context = lambda_context
        self._event_dict = event_dict

    def _base64decode(self, encoded):
        if not isinstance(encoded, bytes):
            encoded = encoded.encode('ascii')
        output = base64.b64decode(encoded)
        return output

    @property
    def raw_body(self):
        if not self._raw_body and self._body is not None:
            if self._is_base64_encoded:
                self._raw_body = self._base64decode(self._body)
            elif not isinstance(self._body, bytes):
                self._raw_body = self._body.encode('utf-8')
            else:
                self._raw_body = self._body
        return self._raw_body

    @property
    def json_body(self):
        if self.headers.get('content-type', '').startswith('application/json'):
            if self._json_body is None:
                try:
                    self._json_body = json.loads(self.raw_body)
                except ValueError:
                    raise BadRequestError('Error Parsing JSON')
            return self._json_body

    def to_dict(self):
        # Don't copy internal attributes.
        copied = {
            k: v for k, v in self.__dict__.items()
            if not k.startswith('_') and
            k not in self._NON_SERIALIZED_ATTRS
        }
        # We want the output of `to_dict()` to be
        # JSON serializable, so we need to remove the CaseInsensitive dict.
        copied['headers'] = dict(copied['headers'])
        if copied['query_params'] is not None:
            copied['query_params'] = dict(copied['query_params'])
        return copied

    def to_original_event(self):
        # To bring consistency with the BaseLambdaEvents, every
        # input event should have access to the original event
        # dictionary as an escape hatch to the underlying data
        # in case something gets added and we haven't mapped it yet.
        # We unfortunately already have a `to_dict()` method which is
        # what other events use so we have to use a different method name.
        return self._event_dict


class Response(object):
    def __init__(self, body, headers=None, status_code=200):
        self.body = body
        if headers is None:
            headers = {}
        self.headers = headers
        self.status_code = status_code

    def to_dict(self, binary_types=None):
        body = self.body
        if not isinstance(body, _ANY_STRING):
            body = json.dumps(body, separators=(',', ':'),
                              default=handle_extra_types)
        single_headers, multi_headers = self._sort_headers(self.headers)
        response = {
            'headers': single_headers,
            'multiValueHeaders': multi_headers,
            'statusCode': self.status_code,
            'body': body
        }
        if binary_types is not None:
            self._b64encode_body_if_needed(response, binary_types)
        return response

    def _sort_headers(self, all_headers):
        multi_headers = {}
        single_headers = {}
        for name, value in all_headers.items():
            if isinstance(value, list):
                multi_headers[name] = value
            else:
                single_headers[name] = value
        return single_headers, multi_headers

    def _b64encode_body_if_needed(self, response_dict, binary_types):
        response_headers = CaseInsensitiveMapping(response_dict['headers'])
        content_type = response_headers.get('content-type', '')
        body = response_dict['body']

        if _matches_content_type(content_type, binary_types):
            if _matches_content_type(content_type, ['application/json']) or \
                    not content_type:
                # There's a special case when a user configures
                # ``application/json`` as a binary type.  The default
                # json serialization results in a string type, but for binary
                # content types we need a type bytes().  So we need to special
                # case this scenario and encode the JSON body to bytes().
                #
                # If a user does not provide a content type header, which can
                # happen if they return a python type instead of a ``Response``
                # type, then we assume the content is application/json.
                body = body if isinstance(body, bytes) \
                    else body.encode('utf-8')
            body = self._base64encode(body)
            response_dict['isBase64Encoded'] = True
        response_dict['body'] = body

    def _base64encode(self, data):
        if not isinstance(data, bytes):
            raise ValueError('Expected bytes type for body with binary '
                             'Content-Type. Got %s type body instead.'
                             % type(data))
        data = base64.b64encode(data)
        return data.decode('ascii')


class RouteEntry(object):

    def __init__(self, view_function, view_name, path, method,
                 api_key_required=None, content_types=None,
                 cors=False, authorizer=None):
        self.view_function = view_function
        self.view_name = view_name
        self.uri_pattern = path
        self.method = method
        self.api_key_required = api_key_required
        #: A list of names to extract from path:
        #: e.g, '/foo/{bar}/{baz}/qux -> ['bar', 'baz']
        self.view_args = self._parse_view_args()
        self.content_types = content_types
        # cors is passed as either a boolean or a CORSConfig object. If it is a
        # boolean it needs to be replaced with a real CORSConfig object to
        # pass the typechecker. None in this context will not inject any cors
        # headers, otherwise the CORSConfig object will determine which
        # headers are injected.
        if cors is True:
            cors = CORSConfig()
        elif cors is False:
            cors = None
        self.cors = cors
        self.authorizer = authorizer

    def _parse_view_args(self):
        if '{' not in self.uri_pattern:
            return []
        # The [1:-1] slice is to remove the braces
        # e.g {foobar} -> foobar
        results = [r[1:-1] for r in _PARAMS.findall(self.uri_pattern)]
        return results

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class APIGateway(object):

    _DEFAULT_BINARY_TYPES = [
        'application/octet-stream', 'application/x-tar', 'application/zip',
        'audio/basic', 'audio/ogg', 'audio/mp4', 'audio/mpeg', 'audio/wav',
        'audio/webm', 'image/png', 'image/jpg', 'image/jpeg', 'image/gif',
        'video/ogg', 'video/mpeg', 'video/webm',
    ]

    def __init__(self):
        self.binary_types = self.default_binary_types
        self.cors = False

    @property
    def default_binary_types(self):
        return list(self._DEFAULT_BINARY_TYPES)


class WebsocketAPI(object):
    _WEBSOCKET_ENDPOINT_TEMPLATE = 'https://{domain_name}/{stage}'
    _REGION_ENV_VARS = ['AWS_REGION', 'AWS_DEFAULT_REGION']

    def __init__(self, env=None):
        self.session = None
        self._endpoint = None
        self._client = None
        if env is None:
            env = os.environ
        self._env = env

    def configure(self, domain_name, stage):
        if self._endpoint is not None:
            return
        self._endpoint = self._WEBSOCKET_ENDPOINT_TEMPLATE.format(
            domain_name=domain_name,
            stage=stage,
        )

    def configure_from_api_id(self, api_id, stage):
        if self._endpoint is not None:
            return
        region_name = self._get_region()

        if region_name.startswith("cn-"):
            domain_name_template = (
                '{api_id}.execute-api.{region}.amazonaws.com.cn'
            )
        else:
            domain_name_template = (
                '{api_id}.execute-api.{region}.amazonaws.com'
            )

        domain_name = domain_name_template.format(
            api_id=api_id, region=region_name)
        self.configure(domain_name, stage)

    def _get_region(self):
        # Attempt to get the region so we can configure the
        # apigatewaymanagementapi client.  We'll first try
        # retrieving this value from env vars because these should
        # always be set in the Lambda runtime environment.
        for varname in self._REGION_ENV_VARS:
            if varname in self._env:
                return self._env[varname]
        # As a last attempt we'll try to retrieve the region
        # from the currently configured region.  If the session
        # isn't configured or we can't get the region, we have
        # no choice but to error out.
        if self.session is not None:
            region_name = self.session.region_name
            if region_name is not None:
                return region_name
        raise ValueError(
            "Unable to retrieve the region name when configuring the "
            "websocket client.  Either set the 'AWS_REGION' environment "
            "variable or assign 'app.websocket_api.session' to a boto3 "
            "session."
        )

    def _get_client(self):
        if self.session is None:
            raise ValueError(
                'Assign app.websocket_api.session to a boto3 session before '
                'using the WebsocketAPI'
            )
        if self._endpoint is None:
            raise ValueError(
                'WebsocketAPI.configure must be called before using the '
                'WebsocketAPI'
            )
        if self._client is None:
            self._client = self.session.client(
                'apigatewaymanagementapi',
                endpoint_url=self._endpoint,
            )
        return self._client

    def send(self, connection_id, message):
        client = self._get_client()
        try:
            client.post_to_connection(
                ConnectionId=connection_id,
                Data=message,
            )
        except client.exceptions.GoneException:
            raise WebsocketDisconnectedError(connection_id)

    def close(self, connection_id):
        client = self._get_client()
        try:
            client.delete_connection(
                ConnectionId=connection_id,
            )
        except client.exceptions.GoneException:
            raise WebsocketDisconnectedError(connection_id)

    def info(self, connection_id):
        client = self._get_client()
        try:
            return client.get_connection(
                ConnectionId=connection_id,
            )
        except client.exceptions.GoneException:
            raise WebsocketDisconnectedError(connection_id)


class DecoratorAPI(object):
    def middleware(self, event_type='all'):
        def _middleware_wrapper(func):
            self.register_middleware(func, event_type)
            return func
        return _middleware_wrapper

    def authorizer(self, ttl_seconds=None, execution_role=None,
                   name=None, header='Authorization'):
        return self._create_registration_function(
            handler_type='authorizer',
            name=name,
            registration_kwargs={
                'ttl_seconds': ttl_seconds,
                'execution_role': execution_role,
                'header': header
            }
        )

    def on_s3_event(self, bucket, events=None,
                    prefix=None, suffix=None, name=None):
        return self._create_registration_function(
            handler_type='on_s3_event',
            name=name,
            registration_kwargs={
                'bucket': bucket, 'events': events,
                'prefix': prefix, 'suffix': suffix,
            }
        )

    def on_sns_message(self, topic, name=None):
        return self._create_registration_function(
            handler_type='on_sns_message',
            name=name,
            registration_kwargs={'topic': topic}
        )

    def on_sqs_message(self, queue=None, batch_size=1,
                       name=None, queue_arn=None,
                       maximum_batching_window_in_seconds=0):
        return self._create_registration_function(
            handler_type='on_sqs_message',
            name=name,
            registration_kwargs={
                'queue': queue,
                'queue_arn': queue_arn,
                'batch_size': batch_size,
                'maximum_batching_window_in_seconds':
                    maximum_batching_window_in_seconds
            }
        )

    def on_cw_event(self, event_pattern, name=None):
        return self._create_registration_function(
            handler_type='on_cw_event',
            name=name,
            registration_kwargs={'event_pattern': event_pattern}
        )

    def schedule(self, expression, name=None, description=''):
        return self._create_registration_function(
            handler_type='schedule',
            name=name,
            registration_kwargs={'expression': expression,
                                 'description': description},
        )

    def on_kinesis_record(self, stream, batch_size=100,
                          starting_position='LATEST', name=None,
                          maximum_batching_window_in_seconds=0):
        return self._create_registration_function(
            handler_type='on_kinesis_record',
            name=name,
            registration_kwargs={
                'stream': stream,
                'batch_size': batch_size,
                'starting_position': starting_position,
                'maximum_batching_window_in_seconds':
                    maximum_batching_window_in_seconds},
        )

    def on_dynamodb_record(self, stream_arn, batch_size=100,
                           starting_position='LATEST', name=None,
                           maximum_batching_window_in_seconds=0):
        return self._create_registration_function(
            handler_type='on_dynamodb_record',
            name=name,
            registration_kwargs={
                'stream_arn': stream_arn,
                'batch_size': batch_size,
                'starting_position': starting_position,
                'maximum_batching_window_in_seconds':
                    maximum_batching_window_in_seconds},
        )

    def route(self, path, **kwargs):
        return self._create_registration_function(
            handler_type='route',
            name=kwargs.pop('name', None),
            # This looks a little weird taking kwargs as a key,
            # but we want to preserve keep the **kwargs signature
            # in the route decorator.
            registration_kwargs={'path': path, 'kwargs': kwargs},
        )

    def lambda_function(self, name=None):
        return self._create_registration_function(
            handler_type='lambda_function', name=name)

    def on_ws_connect(self, name=None):
        return self._create_registration_function(
            handler_type='on_ws_connect',
            name=name,
            registration_kwargs={'route_key': '$connect'},
        )

    def on_ws_disconnect(self, name=None):
        return self._create_registration_function(
            handler_type='on_ws_disconnect',
            name=name,
            registration_kwargs={'route_key': '$disconnect'},
        )

    def on_ws_message(self, name=None):
        return self._create_registration_function(
            handler_type='on_ws_message',
            name=name,
            registration_kwargs={'route_key': '$default'},
        )

    def _create_registration_function(self, handler_type, name=None,
                                      registration_kwargs=None):
        def _register_handler(user_handler):
            handler_name = name
            if handler_name is None:
                handler_name = user_handler.__name__
            if registration_kwargs is not None:
                kwargs = registration_kwargs
            else:
                kwargs = {}
            wrapped = self._wrap_handler(handler_type, handler_name,
                                         user_handler)
            self._register_handler(handler_type, handler_name,
                                   user_handler, wrapped, kwargs)
            return wrapped
        return _register_handler

    def _wrap_handler(self, handler_type, handler_name, user_handler):
        if handler_type in _EVENT_CLASSES:
            if handler_type == 'lambda_function':
                # We have to wrap existing @app.lambda_function()
                # handlers for backwards compat reasons so we can
                # preserve the `def handler(event, context): ...`
                # interface.  However we need a consistent interface
                # for middleware so we have to wrap the event
                # here.
                user_handler = PureLambdaWrapper(user_handler)
            return EventSourceHandler(
                user_handler, _EVENT_CLASSES[handler_type],
                middleware_handlers=self._get_middleware_handlers(
                    event_type=_MIDDLEWARE_MAPPING[handler_type],
                )
            )

        websocket_event_classes = [
            'on_ws_connect',
            'on_ws_message',
            'on_ws_disconnect',
        ]
        if handler_type in websocket_event_classes:
            return WebsocketEventSourceHandler(
                user_handler, WebsocketEvent,
                self.websocket_api,  # pylint: disable=no-member
                middleware_handlers=self._get_middleware_handlers(
                    event_type='websocket')
            )
        if handler_type == 'authorizer':
            # Authorizer is special cased and doesn't quite fit the
            # EventSourceHandler pattern.
            return ChaliceAuthorizer(handler_name, user_handler)
        return user_handler

    def _get_middleware_handlers(self, event_type):
        raise NotImplementedError("_get_middleware_handlers")

    def _register_handler(self, handler_type, name,
                          user_handler, wrapped_handler, kwargs, options=None):
        raise NotImplementedError("_register_handler")

    def register_middleware(self, func, event_type='all'):
        raise NotImplementedError("register_middleware")


class _HandlerRegistration(object):

    def __init__(self):
        self.routes = defaultdict(dict)
        self.websocket_handlers = {}
        self.builtin_auth_handlers = []
        self.event_sources = []
        self.pure_lambda_functions = []
        self.api = APIGateway()
        self.handler_map = {}
        self.middleware_handlers = []

    def register_middleware(self, func, event_type='all'):
        self.middleware_handlers.append((func, event_type))

    def _do_register_handler(self, handler_type, name, user_handler,
                             wrapped_handler, kwargs, options=None):
        url_prefix = None
        name_prefix = None
        module_name = 'app'
        if options is not None:
            name_prefix = options.get('name_prefix')
            if name_prefix is not None:
                name = name_prefix + name
            url_prefix = options.get('url_prefix')
            if url_prefix is not None and handler_type == 'route':
                # Move url_prefix into kwargs so only the
                # route() handler gets a url_prefix kwarg.
                kwargs['url_prefix'] = url_prefix
            # module_name is always provided if options is not None.
            module_name = options['module_name']
        handler_string = '%s.%s' % (module_name, user_handler.__name__)
        getattr(self, '_register_%s' % handler_type)(
            name=name,
            user_handler=user_handler,
            handler_string=handler_string,
            wrapped_handler=wrapped_handler,
            kwargs=kwargs,
        )
        self.handler_map[name] = wrapped_handler

    def _attach_websocket_handler(self, handler):
        route_key = handler.route_key_handled
        decorator_name = {
            '$default': 'on_ws_message',
            '$connect': 'on_ws_connect',
            '$disconnect': 'on_ws_disconnect',
        }.get(route_key)
        if route_key in self.websocket_handlers:
            raise ValueError(
                "Duplicate websocket handler: '%s'. There can only be one "
                "handler for each websocket decorator." % decorator_name
            )
        self.websocket_handlers[route_key] = handler

    def _register_on_ws_connect(self, name, user_handler, handler_string,
                                kwargs, **unused):
        wrapper = WebsocketConnectConfig(
            name=name,
            handler_string=handler_string,
            user_handler=user_handler,
        )
        self._attach_websocket_handler(wrapper)

    def _register_on_ws_message(self, name, user_handler, handler_string,
                                kwargs, **unused):
        route_key = kwargs['route_key']
        wrapper = WebsocketMessageConfig(
            name=name,
            route_key_handled=route_key,
            handler_string=handler_string,
            user_handler=user_handler,
        )
        self._attach_websocket_handler(wrapper)
        self.websocket_handlers[route_key] = wrapper

    def _register_on_ws_disconnect(self, name, user_handler,
                                   handler_string, kwargs, **unused):
        wrapper = WebsocketDisconnectConfig(
            name=name,
            handler_string=handler_string,
            user_handler=user_handler,
        )
        self._attach_websocket_handler(wrapper)

    def _register_lambda_function(self, name, user_handler,
                                  handler_string, **unused):
        wrapper = LambdaFunction(
            user_handler, name=name,
            handler_string=handler_string,
        )
        self.pure_lambda_functions.append(wrapper)

    def _register_on_s3_event(self, name, handler_string, kwargs, **unused):
        events = kwargs['events']
        if events is None:
            events = ['s3:ObjectCreated:*']
        s3_event = S3EventConfig(
            name=name,
            bucket=kwargs['bucket'],
            events=events,
            prefix=kwargs['prefix'],
            suffix=kwargs['suffix'],
            handler_string=handler_string,
        )
        self.event_sources.append(s3_event)

    def _register_on_sns_message(self, name, handler_string, kwargs, **unused):
        sns_config = SNSEventConfig(
            name=name,
            handler_string=handler_string,
            topic=kwargs['topic'],
        )
        self.event_sources.append(sns_config)

    def _register_on_sqs_message(self, name, handler_string, kwargs, **unused):
        queue = kwargs.get('queue')
        queue_arn = kwargs.get('queue_arn')
        if not queue and not queue_arn:
            raise ValueError(
                "Must provide either `queue` or `queue_arn` to the "
                "`on_sqs_message` decorator."
            )
        sqs_config = SQSEventConfig(
            name=name,
            handler_string=handler_string,
            queue=queue,
            queue_arn=queue_arn,
            batch_size=kwargs['batch_size'],
            maximum_batching_window_in_seconds=kwargs[
                'maximum_batching_window_in_seconds'],
        )
        self.event_sources.append(sqs_config)

    def _register_on_kinesis_record(self, name, handler_string,
                                    kwargs, **unused):
        kinesis_config = KinesisEventConfig(
            name=name,
            handler_string=handler_string,
            stream=kwargs['stream'],
            batch_size=kwargs['batch_size'],
            starting_position=kwargs['starting_position'],
            maximum_batching_window_in_seconds=kwargs[
                'maximum_batching_window_in_seconds'],
        )
        self.event_sources.append(kinesis_config)

    def _register_on_dynamodb_record(self, name, handler_string,
                                     kwargs, **unused):
        ddb_config = DynamoDBEventConfig(
            name=name,
            handler_string=handler_string,
            stream_arn=kwargs['stream_arn'],
            batch_size=kwargs['batch_size'],
            starting_position=kwargs['starting_position'],
            maximum_batching_window_in_seconds=kwargs[
                'maximum_batching_window_in_seconds'],
        )
        self.event_sources.append(ddb_config)

    def _register_on_cw_event(self, name, handler_string, kwargs, **unused):
        event_source = CloudWatchEventConfig(
            name=name,
            event_pattern=kwargs['event_pattern'],
            handler_string=handler_string
        )
        self.event_sources.append(event_source)

    def _register_schedule(self, name, handler_string, kwargs, **unused):
        event_source = ScheduledEventConfig(
            name=name,
            schedule_expression=kwargs['expression'],
            description=kwargs["description"],
            handler_string=handler_string,
        )
        self.event_sources.append(event_source)

    def _register_authorizer(self, name, handler_string, wrapped_handler,
                             kwargs, **unused):
        actual_kwargs = kwargs.copy()
        ttl_seconds = actual_kwargs.pop('ttl_seconds', None)
        execution_role = actual_kwargs.pop('execution_role', None)
        header = actual_kwargs.pop('header', None)
        if actual_kwargs:
            raise TypeError(
                'TypeError: authorizer() got unexpected keyword '
                'arguments: %s' % ', '.join(list(actual_kwargs)))
        auth_config = BuiltinAuthConfig(
            name=name,
            handler_string=handler_string,
            ttl_seconds=ttl_seconds,
            execution_role=execution_role,
            header=header,
        )
        wrapped_handler.config = auth_config
        self.builtin_auth_handlers.append(auth_config)

    def _register_route(self, name, user_handler, kwargs, **unused):
        actual_kwargs = kwargs['kwargs']
        path = kwargs['path']
        url_prefix = kwargs.pop('url_prefix', None)
        if url_prefix is not None:
            path = '/'.join([url_prefix.rstrip('/'),
                             path.strip('/')]).rstrip('/')
        methods = actual_kwargs.pop('methods', ['GET'])
        route_kwargs = {
            'authorizer': actual_kwargs.pop('authorizer', None),
            'api_key_required': actual_kwargs.pop('api_key_required', None),
            'content_types': actual_kwargs.pop('content_types',
                                               ['application/json']),
            'cors': actual_kwargs.pop('cors', self.api.cors),
        }
        if route_kwargs['cors'] is None:
            route_kwargs['cors'] = self.api.cors
        if not isinstance(route_kwargs['content_types'], list):
            raise ValueError(
                'In view function "%s", the content_types '
                'value must be a list, not %s: %s' % (
                    name, type(route_kwargs['content_types']),
                    route_kwargs['content_types']))
        if actual_kwargs:
            raise TypeError('TypeError: route() got unexpected keyword '
                            'arguments: %s' % ', '.join(list(actual_kwargs)))
        for method in methods:
            if method in self.routes[path]:
                raise ValueError(
                    "Duplicate method: '%s' detected for route: '%s'\n"
                    "between view functions: \"%s\" and \"%s\". A specific "
                    "method may only be specified once for "
                    "a particular path." % (
                        method, path, self.routes[path][method].view_name,
                        name)
                )
            entry = RouteEntry(user_handler, name, path, method,
                               **route_kwargs)
            self.routes[path][method] = entry


class Chalice(_HandlerRegistration, DecoratorAPI):
    FORMAT_STRING = '%(name)s - %(levelname)s - %(message)s'

    def __init__(self, app_name, debug=False, configure_logs=True, env=None):
        super(Chalice, self).__init__()
        self.app_name = app_name
        self.websocket_api = WebsocketAPI()
        self.current_request = None
        self.lambda_context = None
        self._debug = debug
        self.configure_logs = configure_logs
        self.log = logging.getLogger(self.app_name)
        if env is None:
            env = os.environ
        self._initialize(env)
        self.experimental_feature_flags = set()
        # This is marked as internal but is intended to be used by
        # any code within Chalice.
        self._features_used = set()

    def _initialize(self, env):
        if self.configure_logs:
            self._configure_logging()
        env['AWS_EXECUTION_ENV'] = '%s aws-chalice/%s' % (
            env.get('AWS_EXECUTION_ENV', 'AWS_Lambda'),
            __version__,
        )

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = value
        self._configure_log_level()

    def _configure_logging(self):
        if self._already_configured(self.log):
            return
        handler = logging.StreamHandler(sys.stdout)
        # Timestamp is handled by lambda itself so the
        # default FORMAT_STRING doesn't need to include it.
        formatter = logging.Formatter(self.FORMAT_STRING)
        handler.setFormatter(formatter)
        self.log.propagate = False
        self._configure_log_level()
        self.log.addHandler(handler)

    def _already_configured(self, log):
        if not log.handlers:
            return False
        for handler in log.handlers:
            if isinstance(handler, logging.StreamHandler):
                if handler.stream == sys.stdout:
                    return True
        return False

    def _configure_log_level(self):
        if self._debug:
            level = logging.DEBUG
        else:
            level = logging.ERROR
        self.log.setLevel(level)

    def register_blueprint(self, blueprint, name_prefix=None, url_prefix=None):
        blueprint.register(self, options={'name_prefix': name_prefix,
                                          'url_prefix': url_prefix})

    def _register_handler(self, handler_type, name, user_handler,
                          wrapped_handler, kwargs, options=None):
        self._do_register_handler(handler_type, name, user_handler,
                                  wrapped_handler, kwargs, options)

    # These are defined here on the Chalice class because we want all the
    # feature flag tracking to live in Chalice and not the DecoratorAPI.
    def _register_on_ws_connect(self, name, user_handler, handler_string,
                                kwargs, **unused):
        self._features_used.add('WEBSOCKETS')
        super(Chalice, self)._register_on_ws_connect(
            name, user_handler, handler_string, kwargs, **unused)

    def _register_on_ws_message(self, name, user_handler, handler_string,
                                kwargs, **unused):
        self._features_used.add('WEBSOCKETS')
        super(Chalice, self)._register_on_ws_message(
            name, user_handler, handler_string, kwargs, **unused)

    def _register_on_ws_disconnect(self, name, user_handler,
                                   handler_string, kwargs, **unused):
        self._features_used.add('WEBSOCKETS')
        super(Chalice, self)._register_on_ws_disconnect(
            name, user_handler, handler_string, kwargs, **unused)

    def _get_middleware_handlers(self, event_type):
        # We're returning a generator here because we want to defer the
        # collection of all middleware until as last as possible (when
        # then handler is actually invoked).  This lets us pick up any
        # middleware that's registered after a handler has been defined,
        # which is the behavior you'd expect.
        return (func for func, filter_type in self.middleware_handlers if
                filter_type in [event_type, 'all'])

    def __call__(self, event, context):
        # For legacy reasons, we can't move the Rest API handler entry
        # point away from this Chalice.__call__ method . However, we can
        # try to extract as much as logic as possible to a separate handler
        # class we can call.  That way it's still structured somewhat similar
        # to the other event handlers which makes it more manageable to
        # implement shared functionality (e.g. middleware).
        self.lambda_context = context
        handler = RestAPIEventHandler(
            self.routes, self.api, self.log, self.debug,
            middleware_handlers=self._get_middleware_handlers('http'),
        )
        self.current_request = handler.create_request_object(event, context)
        return handler(event, context)


class BuiltinAuthConfig(object):
    def __init__(self, name, handler_string, ttl_seconds=None,
                 execution_role=None, header='Authorization'):
        # We'd also support all the misc config options you can set.
        self.name = name
        self.handler_string = handler_string
        self.ttl_seconds = ttl_seconds
        self.execution_role = execution_role
        self.header = header


# ChaliceAuthorizer is unique in that the runtime component (the thing
# that wraps the decorated function) also needs a reference to the config
# object (the object the describes how to create the resource).  In
# most event sources these are separate and don't need to know about
# each other, but ChaliceAuthorizer does.  This is because the way
# you associate a builtin authorizer with a view function is by passing
# a direct reference:
#
# @app.authorizer(...)
# def my_auth_function(...): pass
#
# @app.route('/', auth=my_auth_function)
#
# The 'route' part needs to know about the auth function for two reasons:
#
# 1. We use ``view.authorizer`` to figure out how to deploy the app
# 2. We need a reference to the runtime handler for the auth in order
#    to support local mode testing.
# I *think* we can refactor things to handle both of those issues but
# we would need more research to know for sure.  For now, this is a
# special cased runtime class that knows about its config.
class ChaliceAuthorizer(object):
    def __init__(self, name, func, scopes=None):
        self.name = name
        self.func = func
        self.scopes = scopes or []
        # This is filled in during the @app.authorizer()
        # processing.
        self.config = None

    def __call__(self, event, context):
        auth_request = self._transform_event(event)
        result = self.func(auth_request)
        if isinstance(result, AuthResponse):
            return result.to_dict(auth_request)
        return result

    def _transform_event(self, event):
        return AuthRequest(event['type'],
                           event['authorizationToken'],
                           event['methodArn'])

    def with_scopes(self, scopes):
        authorizer_with_scopes = copy.deepcopy(self)
        authorizer_with_scopes.scopes = scopes
        return authorizer_with_scopes


class AuthRequest(object):
    def __init__(self, auth_type, token, method_arn):
        self.auth_type = auth_type
        self.token = token
        self.method_arn = method_arn


class AuthResponse(object):
    ALL_HTTP_METHODS = ['DELETE', 'HEAD', 'OPTIONS',
                        'PATCH', 'POST', 'PUT', 'GET']

    def __init__(self, routes, principal_id, context=None):
        self.routes = routes
        self.principal_id = principal_id
        # The request is used to generate full qualified ARNs
        # that we need for the resource portion of the returned
        # policy.
        if context is None:
            context = {}
        self.context = context

    def to_dict(self, request):
        return {
            'context': self.context,
            'principalId': self.principal_id,
            'policyDocument': self._generate_policy(request),
        }

    def _generate_policy(self, request):
        allowed_resources = self._generate_allowed_resources(request)
        return {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': 'execute-api:Invoke',
                    'Effect': 'Allow',
                    'Resource': allowed_resources,
                }
            ]
        }

    def _generate_allowed_resources(self, request):
        allowed_resources = []
        for route in self.routes:
            if isinstance(route, AuthRoute):
                methods = route.methods
                path = route.path
            elif route == '*':
                # A string route of '*' means that all paths and
                # all HTTP methods are now allowed.
                methods = ['*']
                path = '*'
            else:
                # If 'route' is just a string, then they've
                # opted not to use the AuthRoute(), so we'll
                # generate a policy that allows all HTTP methods.
                methods = ['*']
                path = route
            for method in methods:
                allowed_resources.append(
                    self._generate_arn(path, request, method))
        return allowed_resources

    def _generate_arn(self, route, request, method='*'):
        incoming_arn = request.method_arn
        # An incoming_arn would look like this:
        # "arn:aws:execute-api:us-west-2:123:rest-api-id/stage/GET/needs/auth"
        # Then we pull out the rest-api-id and stage, such that:
        #   base = ['rest-api-id', 'stage']
        #
        # We rely on the fact that the first part of the ARN format is fixed
        # as:    arn:<partition>:<service>:<region>:<account-id>:<resource>
        arn_parts = incoming_arn.split(':', 5)
        allowed_resource = arn_parts[-1].split('/')[:2]
        # Now we add in the path components and rejoin everything
        # back together to make a full arn.
        # We're also assuming all HTTP methods (via '*') for now.
        # To support per HTTP method routes the API will need to be updated.
        # We also need to strip off the leading ``/`` so it can be
        # '/'.join(...)'d properly.
        allowed_resource.extend([method, route[1:]])
        last_arn_segment = '/'.join(allowed_resource)
        if route == '*':
            # We also have to handle the '*' case which matches
            # all routes.
            last_arn_segment += route
        arn_parts[-1] = last_arn_segment
        final_arn = ':'.join(arn_parts)
        return final_arn


class AuthRoute(object):
    def __init__(self, path, methods):
        self.path = path
        self.methods = methods


class LambdaFunction(object):
    def __init__(self, func, name, handler_string):
        self.func = func
        self.name = name
        self.handler_string = handler_string

    def __call__(self, event, context):
        return self.func(event, context)


class BaseEventSourceConfig(object):
    def __init__(self, name, handler_string):
        self.name = name
        self.handler_string = handler_string


class ScheduledEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, schedule_expression, description):
        super(ScheduledEventConfig, self).__init__(name, handler_string)
        self.schedule_expression = schedule_expression
        self.description = description


class CloudWatchEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, event_pattern):
        super(CloudWatchEventConfig, self).__init__(name, handler_string)
        self.event_pattern = event_pattern


class ScheduleExpression(object):
    def to_string(self):
        raise NotImplementedError("to_string")


class Rate(ScheduleExpression):
    MINUTES = 'MINUTES'
    HOURS = 'HOURS'
    DAYS = 'DAYS'

    def __init__(self, value, unit):
        self.value = value
        self.unit = unit

    def to_string(self):
        unit = self.unit.lower()
        if self.value == 1:
            # Remove the 's' from the end if it's singular.
            # This is required by the cloudwatch events API.
            unit = unit[:-1]
        return 'rate(%s %s)' % (self.value, unit)


class Cron(ScheduleExpression):
    def __init__(self, minutes, hours, day_of_month, month, day_of_week, year):
        self.minutes = minutes
        self.hours = hours
        self.day_of_month = day_of_month
        self.month = month
        self.day_of_week = day_of_week
        self.year = year

    def to_string(self):
        return 'cron(%s %s %s %s %s %s)' % (
            self.minutes,
            self.hours,
            self.day_of_month,
            self.month,
            self.day_of_week,
            self.year,
        )


class S3EventConfig(BaseEventSourceConfig):
    def __init__(self, name, bucket, events, prefix, suffix, handler_string):
        super(S3EventConfig, self).__init__(name, handler_string)
        self.bucket = bucket
        self.events = events
        self.prefix = prefix
        self.suffix = suffix


class SNSEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, topic):
        super(SNSEventConfig, self).__init__(name, handler_string)
        self.topic = topic


class SQSEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, queue, queue_arn, batch_size,
                 maximum_batching_window_in_seconds):
        super(SQSEventConfig, self).__init__(name, handler_string)
        self.queue = queue
        self.queue_arn = queue_arn
        self.batch_size = batch_size
        self.maximum_batching_window_in_seconds = \
            maximum_batching_window_in_seconds


class KinesisEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, stream,
                 batch_size, starting_position,
                 maximum_batching_window_in_seconds):
        super(KinesisEventConfig, self).__init__(name, handler_string)
        self.stream = stream
        self.batch_size = batch_size
        self.starting_position = starting_position
        self.maximum_batching_window_in_seconds = \
            maximum_batching_window_in_seconds


class DynamoDBEventConfig(BaseEventSourceConfig):
    def __init__(self, name, handler_string, stream_arn,
                 batch_size, starting_position,
                 maximum_batching_window_in_seconds):
        super(DynamoDBEventConfig, self).__init__(name, handler_string)
        self.stream_arn = stream_arn
        self.batch_size = batch_size
        self.starting_position = starting_position
        self.maximum_batching_window_in_seconds = \
            maximum_batching_window_in_seconds


class WebsocketConnectConfig(BaseEventSourceConfig):
    CONNECT_ROUTE = '$connect'

    def __init__(self, name, handler_string, user_handler):
        super(WebsocketConnectConfig, self).__init__(name, handler_string)
        self.route_key_handled = self.CONNECT_ROUTE
        self.handler_function = user_handler


class WebsocketMessageConfig(BaseEventSourceConfig):
    def __init__(self, name, route_key_handled, handler_string, user_handler):
        super(WebsocketMessageConfig, self).__init__(name, handler_string)
        self.route_key_handled = route_key_handled
        self.handler_function = user_handler


class WebsocketDisconnectConfig(BaseEventSourceConfig):
    DISCONNECT_ROUTE = '$disconnect'

    def __init__(self, name, handler_string, user_handler):
        super(WebsocketDisconnectConfig, self).__init__(name, handler_string)
        self.route_key_handled = self.DISCONNECT_ROUTE
        self.handler_function = user_handler


class PureLambdaWrapper(object):
    def __init__(self, original_func):
        self._original_func = original_func

    def __call__(self, event):
        # The @app.lambda_function() expects an event dict
        # and a context argument so this class will is used to adapt
        # from the Chalice single-arg style function (which is used
        # in all the event handlers) to the low-level lambda api.
        return self._original_func(event.to_dict(), event.context)


class MiddlewareHandler(object):
    def __init__(self, handler, next_handler):
        self.handler = handler
        self.next_handler = next_handler

    def __call__(self, request):
        return self.handler(request, self.next_handler)


class BaseLambdaHandler(object):
    def __call__(self, event, context):
        pass

    def _build_middleware_handlers(self, handlers, original_handler):
        current = original_handler
        for handler in reversed(list(handlers)):
            current = MiddlewareHandler(handler=handler, next_handler=current)
        return current


class EventSourceHandler(BaseLambdaHandler):

    def __init__(self, func, event_class, middleware_handlers=None):
        self.func = func
        self.event_class = event_class
        if middleware_handlers is None:
            middleware_handlers = []
        self._middleware_handlers = middleware_handlers
        self.handler = None

    @property
    def middleware_handlers(self):
        return self._middleware_handlers

    @middleware_handlers.setter
    def middleware_handlers(self, value):
        self._middleware_handlers = value

    def __call__(self, event, context):
        event_obj = self.event_class(event, context)
        if self.handler is None:
            # Defer creating handlers so we have all middleware configured.
            self.handler = self._build_middleware_handlers(
                self._middleware_handlers, original_handler=self.func)
        return self.handler(event_obj)


class WebsocketEventSourceHandler(EventSourceHandler):
    WEBSOCKET_API_RESPONSE = {'statusCode': 200}

    def __init__(self, func, event_class, websocket_api,
                 middleware_handlers=None):
        self.func = func
        self.event_class = event_class
        self.websocket_api = websocket_api
        if middleware_handlers is None:
            middleware_handlers = []
        self._middleware_handlers = middleware_handlers
        self.handler = None

    def __call__(self, event, context):
        self.websocket_api.configure_from_api_id(
            event['requestContext']['apiId'],
            event['requestContext']['stage'],
        )
        response = super(
            WebsocketEventSourceHandler, self).__call__(event, context)
        data = None
        if isinstance(response, Response):
            data = response.to_dict()
        elif isinstance(response, dict):
            data = response
            if "statusCode" not in data:
                data = {**self.WEBSOCKET_API_RESPONSE, **data}
        return data or self.WEBSOCKET_API_RESPONSE


class RestAPIEventHandler(BaseLambdaHandler):
    def __init__(self, route_table, api, log, debug, middleware_handlers=None):
        self.routes = route_table
        self.api = api
        self.log = log
        self.debug = debug
        self.current_request = None
        self.lambda_context = None
        if middleware_handlers is None:
            middleware_handlers = []
        self._middleware_handlers = middleware_handlers

    def _global_error_handler(self, event, get_response):
        try:
            return get_response(event)
        except Exception:
            return self._unhandled_exception_to_response()

    def create_request_object(self, event, context):
        # For legacy reasons, there's some initial validation that takes
        # place before we convert the input event to a python object.
        # We don't do this in event handlers we added later, so we *should*
        # be able to remove this code.  To be safe, we're keeping it in for
        # now to minimize the potential for breaking changes.
        resource_path = event.get('requestContext', {}).get('resourcePath')
        if resource_path is not None:
            self.current_request = Request(event, context)
            return self.current_request

    def __call__(self, event, context):
        def wrapped_event(request):
            return self._main_rest_api_handler(event, context)

        final_handler = self._build_middleware_handlers(
            [self._global_error_handler] + list(self._middleware_handlers),
            original_handler=wrapped_event,
        )
        response = final_handler(self.current_request)
        return response.to_dict(self.api.binary_types)

    def _main_rest_api_handler(self, event, context):
        resource_path = event.get('requestContext', {}).get('resourcePath')
        if resource_path is None:
            return error_response(error_code='InternalServerError',
                                  message='Unknown request.',
                                  http_status_code=500)
        http_method = event['requestContext']['httpMethod']
        if http_method not in self.routes[resource_path]:
            allowed_methods = ', '.join(self.routes[resource_path].keys())
            return error_response(
                error_code='MethodNotAllowedError',
                message='Unsupported method: %s' % http_method,
                http_status_code=405,
                headers={'Allow': allowed_methods})
        route_entry = self.routes[resource_path][http_method]
        view_function = route_entry.view_function
        function_args = {name: event['pathParameters'][name]
                         for name in route_entry.view_args}
        self.lambda_context = context
        # We're getting the CORS headers before validation to be able to
        # output desired headers with
        cors_headers = None
        if self._cors_enabled_for_route(route_entry):
            cors_headers = self._get_cors_headers(route_entry.cors)
        # We're doing the header validation after creating the request
        # so can leverage the case insensitive dict that the Request class
        # uses for headers.
        if route_entry.content_types:
            content_type = self.current_request.headers.get(
                'content-type', 'application/json')
            if not _matches_content_type(content_type,
                                         route_entry.content_types):
                return error_response(
                    error_code='UnsupportedMediaType',
                    message='Unsupported media type: %s' % content_type,
                    http_status_code=415,
                    headers=cors_headers
                )
        response = self._get_view_function_response(view_function,
                                                    function_args)
        if cors_headers is not None:
            self._add_cors_headers(response, cors_headers)

        response_headers = CaseInsensitiveMapping(response.headers)
        if not self._validate_binary_response(
                self.current_request.headers, response_headers):
            content_type = response_headers.get('content-type', '')
            return error_response(
                error_code='BadRequest',
                message=('Request did not specify an Accept header with %s, '
                         'The response has a Content-Type of %s. If a '
                         'response has a binary Content-Type then the request '
                         'must specify an Accept header that matches.'
                         % (content_type, content_type)),
                http_status_code=400,
                headers=cors_headers
            )
        return response

    def _validate_binary_response(self, request_headers, response_headers):
        # Validates that a response is valid given the request. If the response
        # content-type specifies a binary type, there must be an accept header
        # that is a binary type as well.
        request_accept_header = request_headers.get('accept')
        response_content_type = response_headers.get(
            'content-type', 'application/json')
        response_is_binary = _matches_content_type(response_content_type,
                                                   self.api.binary_types)
        expects_binary_response = False
        if request_accept_header is not None:
            expects_binary_response = _matches_content_type(
                request_accept_header, self.api.binary_types)
        if response_is_binary and not expects_binary_response:
            return False
        return True

    def _get_view_function_response(self, view_function, function_args):
        try:
            response = view_function(**function_args)
            if not isinstance(response, Response):
                response = Response(body=response)
            self._validate_response(response)
        except ChaliceUnhandledError:
            # Reraise this exception so that middleware has a chance
            # to handle the exception.
            raise
        except ChaliceViewError as e:
            # Any chalice view error should propagate.  These
            # get mapped to various HTTP status codes in API Gateway.
            response = Response(body={'Code': e.__class__.__name__,
                                      'Message': str(e)},
                                status_code=e.STATUS_CODE)
        except Exception:
            response = self._unhandled_exception_to_response()
        return response

    def _unhandled_exception_to_response(self):
        headers = {}
        path = getattr(self.current_request, 'path', 'unknown')
        self.log.error("Caught exception for path %s", path, exc_info=True)
        if self.debug:
            # If the user has turned on debug mode,
            # we'll let the original exception propagate so
            # they get more information about what went wrong.
            stack_trace = ''.join(traceback.format_exc())
            body = stack_trace
            headers['Content-Type'] = 'text/plain'
        else:
            body = {'Code': 'InternalServerError',
                    'Message': 'An internal server error occurred.'}
        response = Response(body=body, headers=headers, status_code=500)
        return response

    def _validate_response(self, response):
        for header, value in response.headers.items():
            if '\n' in value:
                raise ChaliceError("Bad value for header '%s': %r" %
                                   (header, value))

    def _cors_enabled_for_route(self, route_entry):
        return route_entry.cors is not None

    def _get_cors_headers(self, cors):
        return cors.get_access_control_headers()

    def _add_cors_headers(self, response, cors_headers):
        for name, value in cors_headers.items():
            if name not in response.headers:
                response.headers[name] = value


# These classes contain all the event types that are passed
# in as arguments in the lambda event handlers.  These are
# part of Chalice's public API and must be backwards compatible.

class BaseLambdaEvent(object):
    def __init__(self, event_dict, context):
        self._event_dict = event_dict
        self.context = context
        self._extract_attributes(event_dict)

    def _extract_attributes(self, event_dict):
        raise NotImplementedError("_extract_attributes")

    def to_dict(self):
        return self._event_dict


# This class is only used for middleware handlers because
# we can't change the existing interface for @app.lambda_function().
# This could be a Chalice 2.0 thing where we make all the decorators
# have a consistent interface that takes a single event arg.
class LambdaFunctionEvent(BaseLambdaEvent):
    def __init__(self, event_dict, context):
        self.event = event_dict
        self.context = context

    def _extract_attributes(self, event_dict):
        pass

    def to_dict(self):
        return self.event


class CloudWatchEvent(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        self.version = event_dict['version']
        self.account = event_dict['account']
        self.region = event_dict['region']
        self.detail = event_dict['detail']
        self.detail_type = event_dict['detail-type']
        self.source = event_dict['source']
        self.time = event_dict['time']
        self.event_id = event_dict['id']
        self.resources = event_dict['resources']


class WebsocketEvent(BaseLambdaEvent):
    def __init__(self, event_dict, context):
        super(WebsocketEvent, self).__init__(event_dict, context)
        self._json_body = None

    def _extract_attributes(self, event_dict):
        request_context = event_dict['requestContext']
        self.domain_name = request_context['domainName']
        self.stage = request_context['stage']
        self.connection_id = request_context['connectionId']
        self.body = event_dict.get('body')

    @property
    def json_body(self):
        if self._json_body is None:
            try:
                self._json_body = json.loads(self.body)
            except ValueError:
                raise BadRequestError('Error Parsing JSON')
        return self._json_body


class SNSEvent(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        first_record = event_dict['Records'][0]
        self.message = first_record['Sns']['Message']
        self.subject = first_record['Sns']['Subject']


class S3Event(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        s3 = event_dict['Records'][0]['s3']
        self.bucket = s3['bucket']['name']
        self.key = unquote_str(s3['object']['key'])


class SQSEvent(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        # We don't extract anything off the top level
        # event.
        pass

    def __iter__(self):
        for record in self._event_dict['Records']:
            yield SQSRecord(record, self.context)


class SQSRecord(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        self.body = event_dict['body']
        self.receipt_handle = event_dict['receiptHandle']


class KinesisEvent(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        pass

    def __iter__(self):
        for record in self._event_dict['Records']:
            yield KinesisRecord(record, self.context)


class KinesisRecord(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        kinesis = event_dict['kinesis']
        encoded_payload = kinesis['data']
        self.data = base64.b64decode(encoded_payload)
        self.sequence_number = kinesis['sequenceNumber']
        self.partition_key = kinesis['partitionKey']
        self.schema_version = kinesis['kinesisSchemaVersion']
        self.timestamp = datetime.datetime.utcfromtimestamp(
            kinesis['approximateArrivalTimestamp'])


class DynamoDBEvent(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        pass

    def __iter__(self):
        for record in self._event_dict['Records']:
            yield DynamoDBRecord(record, self.context)


class DynamoDBRecord(BaseLambdaEvent):
    def _extract_attributes(self, event_dict):
        dynamodb = event_dict['dynamodb']
        self.timestamp = datetime.datetime.utcfromtimestamp(
            dynamodb['ApproximateCreationDateTime'])
        self.keys = dynamodb.get('Keys')
        self.new_image = dynamodb.get('NewImage')
        self.old_image = dynamodb.get('OldImage')
        self.sequence_number = dynamodb['SequenceNumber']
        self.size_bytes = dynamodb['SizeBytes']
        self.stream_view_type = dynamodb['StreamViewType']
        # These are from the top level keys in a record.
        self.aws_region = event_dict['awsRegion']
        self.event_id = event_dict['eventID']
        self.event_name = event_dict['eventName']
        self.event_source_arn = event_dict['eventSourceARN']

    @property
    def table_name(self):
        # Converts:
        # "arn:aws:dynamodb:us-west-2:12345:table/MyTable/"
        # "stream/2020-09-28T16:49:14.209"
        #
        # into:
        # "MyTable"
        parts = self.event_source_arn.split(':', 5)
        if not len(parts) == 6:
            return ''
        full_name = parts[-1]
        name_parts = full_name.split('/')
        if len(name_parts) >= 2:
            return name_parts[1]
        return ''


class Blueprint(DecoratorAPI):
    def __init__(self, import_name):
        self._import_name = import_name
        self._deferred_registrations = []
        self._current_app = None
        self._lambda_context = None

    @property
    def log(self):
        if self._current_app is None:
            raise RuntimeError(
                "Can only access Blueprint.log if it's registered to an app."
            )
        return self._current_app.log

    @property
    def current_request(self):
        if self._current_app is None:
            raise RuntimeError(
                "Can only access Blueprint.current_request if it's registered "
                "to an app."
            )
        return self._current_app.current_request

    @property
    def current_app(self):
        if self._current_app is None:
            raise RuntimeError(
                "Can only access Blueprint.current_app if it's registered "
                "to an app."
            )
        return self._current_app

    @property
    def lambda_context(self):
        if self._current_app is None:
            raise RuntimeError(
                "Can only access Blueprint.lambda_context if it's registered "
                "to an app."
            )
        return self._current_app.lambda_context

    def register(self, app, options):
        self._current_app = app
        all_options = options.copy()
        all_options['module_name'] = self._import_name
        for function in self._deferred_registrations:
            function(app, all_options)

    # Note on blueprints implementation.  One option we have for implementing
    # blueprints is to copy every decorator in our public API over to the
    # Blueprints class.  Instead what we do is inherit from DecoratorAPI so we
    # get new decorators for free.  The tradeoff is that need to add
    # implementations of the internal methods used to manage handler
    # registration that defer registration until we get an app object. While
    # these methods are not public in the sense that we don't want users to
    # call them, they're available for blueprints to use in order to avoid
    # boilerplate code.

    def register_middleware(self, func, event_type='all'):
        self._deferred_registrations.append(
            # pylint: disable=protected-access
            lambda app, options: app.register_middleware(
                func, event_type
            )
        )

    def _register_handler(self, handler_type, name, user_handler,
                          wrapped_handler, kwargs, options=None):
        # If we go through the public API (app.route, app.schedule, etc) then
        # we have to duplicate either the methods or the params in this
        # class.  We're using _register_handler as a tradeoff for cutting
        # down on the duplication.
        def _register_blueprint_handler(app, options):
            if handler_type in _EVENT_CLASSES:
                # pylint: disable=protected-access
                wrapped_handler.middleware_handlers = \
                    app._get_middleware_handlers(
                        _MIDDLEWARE_MAPPING[handler_type])
            # pylint: disable=protected-access
            app._register_handler(
                handler_type, name, user_handler, wrapped_handler,
                kwargs, options
            )
        self._deferred_registrations.append(_register_blueprint_handler)

    def _get_middleware_handlers(self, event_type):
        # This will get filled in later during the registration process.
        return []


# This class is used to convert any existing/3rd party decorators
# that work directly on lambda functions with the original signature
# of (event, context).  By using ConvertToMiddleware you can automatically
# apply this decorator to every lambda function in a Chalice app.
# Example:
#
# Before:
#
# @third_part.decorator
# def some_lambda_function(event, context): pass
#
# Now:
#
# app.register_middleware(ConvertToMiddleware(third_party.decorator))
#
#
class ConvertToMiddleware(object):
    def __init__(self, lambda_wrapper):
        self._wrapper = lambda_wrapper

    def __call__(self, event, get_response):
        original_event, context = self._extract_original_param(event)

        @functools.wraps(self._wrapper)
        def wrapped(original_event, context):
            return get_response(event)
        return self._wrapper(wrapped)(original_event, context)

    def _extract_original_param(self, event):
        if isinstance(event, Request):
            return event.to_original_event(), event.lambda_context
        return event.to_dict(), event.context


_EVENT_CLASSES = {
    'on_s3_event': S3Event,
    'on_sns_message': SNSEvent,
    'on_sqs_message': SQSEvent,
    'on_cw_event': CloudWatchEvent,
    'on_kinesis_record': KinesisEvent,
    'on_dynamodb_record': DynamoDBEvent,
    'schedule': CloudWatchEvent,
    'lambda_function': LambdaFunctionEvent,
}


_MIDDLEWARE_MAPPING = {
    'on_s3_event': 's3',
    'on_sns_message': 'sns',
    'on_sqs_message': 'sqs',
    'on_cw_event': 'cloudwatch',
    'on_kinesis_record': 'kinesis',
    'on_dynamodb_record': 'dynamodb',
    'schedule': 'scheduled',
    'lambda_function': 'pure_lambda',
}
